# Copyright (c) 2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilies for filtering a `Table` using column slice definitions."""

from __future__ import annotations

import operator
import re
import token
from io import StringIO
from tokenize import generate_tokens
from typing import (
    TYPE_CHECKING,
    Generic,
    NamedTuple,
    TypeVar,
    cast,
)

import numpy

FilterOperandType = TypeVar("FilterOperandType")

if TYPE_CHECKING:
    from collections.abc import (
        Callable,
        Iterable,
        Iterator,
    )

    from astropy.table import Column

    from . import Table

    T = TypeVar("T", bound=Table)

    FilterTuple = tuple[
        str | tuple[str, ...],
        Callable[[Table | Column, FilterOperandType], numpy.ndarray],
        FilterOperandType,
    ]
    FilterLike = str | FilterTuple

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

OPERATORS: dict[str, Callable] = {
    "<": operator.lt,
    "<=": operator.le,
    "=": operator.eq,
    "==": operator.eq,
    ">=": operator.ge,
    ">": operator.gt,
    "!=": operator.ne,
}

OPERATORS_INV: dict[str, Callable] = {
    "<=": operator.ge,
    "<": operator.gt,
    ">": operator.lt,
    ">=": operator.le,
}

QUOTE_REGEX = re.compile(r"^[\s\"\']+|[\s\"\']+$")
DELIM_REGEX = re.compile(r"(and|&+)", re.IGNORECASE)


# -- filter parsing ------------------

FILTER_SIMPLE_ARG_COUNT = 3
FILTER_COMPOUND_ARG_COUNT = 5

class FilterSpec(NamedTuple, Generic[FilterOperandType]):
    """A table column filter definition.

    The generic type parameter _O represents the type of the operand,
    which should match the second parameter of the operator callable.
    """

    column: str | tuple[str, ...]
    operator: Callable[[Table | Column, FilterOperandType], numpy.ndarray]
    operand: FilterOperandType


def _float_or_str(value: str) -> float | str:
    """Attempt `float(value)` handling a `ValueError`."""
    # remove any surrounding quotes
    value = QUOTE_REGEX.sub("", value)
    try:  # attempt `float()` conversion
        return float(value)
    except ValueError:  # just return the input
        return value


def parse_operator(mathstr: str) -> Callable:
    """Parse a `str` as a function from the `operator` module.

    Parameters
    ----------
    mathstr : `str`
        A `str` representing a mathematical operator.

    Returns
    -------
    op : `func`
        One of the functions from the `operator` module.

    Raises
    ------
    KeyError
        If input `str` cannot be mapped to an `operator` function.

    Examples
    --------
    >>> parse_operator('>')
    <built-in function gt>
    """
    try:
        return OPERATORS[mathstr]
    except KeyError as exc:
        exc.args = (
            f"Unrecognised operator '{mathstr}'",
        )
        raise


def parse_column_filter(definition: str) -> list[FilterSpec]:
    """Parse a `str` of the form 'column>50'.

    Parameters
    ----------
    definition : `str`
        a column filter definition of the form ``<name><operator><threshold>``
        or ``<threshold><operator><name><operator><threshold>``, e.g.
        ``frequency >= 10``, or ``50 < snr < 100``

    Returns
    -------
    filters : `list` of `FilterSpec`
        A `list` of filter 3-`tuple`s, where each `tuple` contains the
        following elements:

        - ``column`` (`str`) - the name of the column on which to operate
        - ``operator`` (`callable`) - the operator to call when evaluating
          the filter
        - ``operand`` (anything) - the argument to the operator function

    Raises
    ------
    ValueError
        If the filter definition cannot be parsed.

    KeyError
        If any parsed operator string cannnot be mapped to a function from
        the `operator` module.

    Notes
    -----
    Strings that contain non-alphanumeric characters (e.g. hyphen `-`) should
    be quoted inside the filter definition, to prevent such characters
    being interpreted as operators, e.g. ``channel = X1:TEST`` should always
    be passed as ``channel = "X1:TEST"``.

    Examples
    --------
    >>> parse_column_filter("frequency>10")
    FilterSpec(column='frequency', operator=<built-in function gt>, operand=10.0)
    >>> parse_column_filter("50 < snr < 100")
    [FilterSpec(column='snr', operator=<built-in function gt>, operand=50.0),
     FilterSpec(column='snr', operator=<built-in function lt>, operand=100.0)]
    >>> parse_column_filter('channel = "H1:TEST"')
    FilterSpec(column='channel', operator=<built-in function eq>, operand='H1:TEST')
    """
    # parse definition into parts (skipping null tokens)
    parts = list(generate_tokens(StringIO(definition.strip()).readline))
    while parts[-1].type in {token.ENDMARKER, token.NEWLINE}:
        parts = parts[:-1]

    # parse simple definition: e.g: snr > 5 or 5 < snr
    if len(parts) == FILTER_SIMPLE_ARG_COUNT:
        left, op_, right = parts
        if left.type in {token.NAME, token.STRING}:  # snr > 5
            name = QUOTE_REGEX.sub("", left.string)
            oprtr = OPERATORS[op_.string]
            value = _float_or_str(right.string)
            return [FilterSpec(name, oprtr, value)]
        if right.type in {token.NAME, token.STRING}:  # 5 < snr
            name = QUOTE_REGEX.sub("", right.string)
            oprtr = OPERATORS_INV[op_.string]
            value = _float_or_str(left.string)
            return [FilterSpec(name, oprtr, value)]

    # parse between definition: e.g: 5 < snr < 10
    elif len(parts) == FILTER_COMPOUND_ARG_COUNT:
        left, op1, mid, op2, right = parts
        name = QUOTE_REGEX.sub("", mid.string)
        return [
            FilterSpec(
                name,
                OPERATORS_INV[op1.string],
                _float_or_str(left.string),
            ),
            FilterSpec(
                name,
                OPERATORS[op2.string],
                _float_or_str(right.string),
            ),
        ]

    msg = f"cannot parse filter definition from '{definition}'"
    raise ValueError(msg)


def parse_column_filters(
    *definitions: FilterLike | Iterable[FilterLike],
) -> list[FilterSpec]:
    """Parse multiple compound column filter definitions.

    Examples
    --------
    >>> parse_column_filters('snr > 10', 'frequency < 1000')
    [FilterSpec(column='snr', operator=<built-in function gt>, operand=10.0),
     FilterSpec(column='frequency', operator=<built-in function lt>, operand=1000.0)]
    >>> parse_column_filters('snr > 10 && frequency < 1000')
    [FilterSpec(column='snr', operator=<built-in function gt>, operand=10.0),
     FilterSpec(column='frequency', operator=<built-in function lt>, operand=1000.0)]
    """
    fltrs = []
    for def_ in _flatten(definitions):
        if isinstance(def_, str):
            for splitdef in DELIM_REGEX.split(def_)[::2]:
                fltrs.extend(parse_column_filter(splitdef))
        else:
            fltrs.append(FilterSpec(*def_))
    return fltrs


def _flatten(
    container: Iterable,
) -> Iterator[FilterLike]:
    """Flatten arbitrary nested list of filters into a 1-D list."""
    for elem in container:
        if isinstance(elem, str) or is_filter_tuple(elem):
            yield elem
        else:
            yield from _flatten(elem)


def is_filter_tuple(tup: object) -> bool:
    """Return whether a `tuple` matches the format for a column filter."""
    if isinstance(tup, FilterSpec):
        return True
    try:
        names, func, _ = cast("tuple[str | tuple[str, ...], Callable, tuple]", tup)  # type: ignore[misc]
        return (
            (isinstance(names, str) or all(isinstance(x, str) for x in names))
            and callable(func)
        )
    except (TypeError, ValueError):
        return False


# -- filter --------------------------

def filter_table(
    table: T,
    *column_filters: FilterLike | Iterable[FilterLike],
) -> T:
    """Apply one or more column slice filters to a `Table`.

    Multiple column filters can be given, and will be applied serially.

    Parameters
    ----------
    table : `~astropy.table.Table`
        The table to filter.

    column_filters : `str`, `tuple`
        One or more column slice filter definition, in one of two formats:

        - `str` - e.g. ``'snr > 10``
        - `FilterSpec` (`tuple`) - ``(<column(s)>, <operator>, <operand>)``,
          e.g. ``('snr', operator.gt, 10)``

        Multiple filters can be given and will be applied in order.

    Returns
    -------
    table : `~astropy.table.Table`
        A view of the input table with only those rows matching the filters.

    Examples
    --------
    >>> filter(my_table, 'snr>10', 'frequency<1000')

    Custom operations can be defined using filter tuple definitions:

    >>> from gwpy.table.filters import in_segmentlist
    >>> filter(my_table, ('time', in_segmentlist, segs))
    """
    keep = numpy.ones(len(table), dtype=bool)
    for name, op_func, operand in parse_column_filters(*column_filters):
        col = table[name]
        keep &= op_func(col, operand)
    return table[keep]



# Copyright (c) 2024-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unified I/O read/write for `Spectrogram` objects."""

from __future__ import annotations

from ..types.connect import (
    Array2DRead,
    Array2DWrite,
)

# -- Spectrogram -----------------

class SpectrogramRead(Array2DRead):
    """Read data into a `Spectrogram`.

    Arguments and keywords depend on the output format, see the
    online documentation for full details for each format, the
    parameters below are common to most formats.

    Parameters
    ----------
    source : `str`, `list`
        Source of data, any of the following:

        - `str` path of single data file,
        - `str` path of LAL-format cache file,
        - `list` of paths.

    *args
        Other arguments are (in general) specific to the given
        ``format``.

    format : `str`, optional
        Source format identifier. If not given, the format will be
        detected if possible. See below for list of acceptable
        formats.

    **kwargs
        Other keywords are (in general) specific to the given ``format``.

    Raises
    ------
    IndexError
        if ``source`` is an empty list

    Notes
    -----"""


class SpectrogramWrite(Array2DWrite):
    """Write this `Spectrogram` to a file.

    Arguments and keywords depend on the output format, see the
    online documentation for full details for each format, the
    parameters below are common to most formats.

    Parameters
    ----------
    target : `str`
        output filename

    format : `str`, optional
        output format identifier. If not given, the format will be
        detected if possible. See below for list of acceptable
        formats.

    Notes
    -----"""



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read events from an Omicron-format ROOT file."""

from __future__ import annotations

from typing import TYPE_CHECKING

from astropy.table import Table

from .. import EventTable

if TYPE_CHECKING:
    from ...io.utils import Readable

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def table_from_omicron(
    source: Readable,
    treename: str = "triggers",
    **kwargs,
) -> Table:
    """Read an `EventTable` from an Omicron ROOT file.

    This function just redirects to the format='root' reader with appropriate
    defaults.

    See `EventTable.read.help('root')` for more details.
    """
    return Table.read(
        source,
        format="root",
        treename=treename,
        **kwargs,
    )


for _klass in (Table, EventTable):
    _klass.read.registry.register_reader(
        "root.omicron",
        _klass,
        table_from_omicron,
    )



# Copyright (c) 2023-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Custom filtering utilities for the `TimeSeries`."""

__author__ = "Alex Southgate <alex.southgate@ligo.org>"

import warnings

import numpy as np
import pytest
import scipy.signal as sig

from ...signal import spectral
from ...timeseries import TimeSeries

RNG = np.random.default_rng(12345)


@pytest.fixture
def series_data():
    """Create some fake data with equal sampling frequencies.

    Returns
    -------
        firstarr: an array of data, simple mixture of waves
        secondarr: a second array of data from different mixture
        seglen: segment length param to reuse for ffts
    """
    seglen = 512
    n_segs = 10
    n_t = seglen * n_segs
    t_end = 2 * np.pi

    ts = np.linspace(0, t_end, n_t)

    firstarr = 0.1 * np.cos(ts + 0.1) + 0.9 * np.sin(2 * ts + 5)
    firstarr += RNG.normal(5.8, 2, n_t)

    secondarr = 0.5 * np.cos(ts + 0.1) + 0.1 * np.sin(5 * ts + 10)
    secondarr += RNG.normal(5.8, 2, n_t)

    return firstarr, secondarr, seglen


def test_coherence_happy(series_data):
    """Test the interface to scipy.signal.coherence.

    For other tests see timeseries/tests/timeseries.py
    """
    firstarr, secondarr, seglen = series_data
    f_s = 0.001

    first = TimeSeries(firstarr, sample_rate=f_s)
    second = TimeSeries(secondarr, sample_rate=f_s)

    coh = spectral.coherence(first, second, segmentlength=seglen)
    _, cxytemp = sig.coherence(firstarr, secondarr, f_s, nperseg=seglen)

    coharr = np.array(coh.data)

    assert all(coharr == cxytemp)


def test_coherence_resample(series_data):
    """Ensure warning is raised by unequal sampling frequencies."""
    firstarr, secondarr, seglen = series_data
    f_s = 0.001

    first = TimeSeries(firstarr, sample_rate=f_s)
    second = TimeSeries(secondarr, sample_rate=f_s * 2.32)

    with pytest.warns(
        UserWarning,
        match="Sampling frequencies are unequal",
    ):
        coh1 = spectral.coherence(first, second, segmentlength=seglen)

    # check that forcibly disabling downsample results in an error
    with pytest.raises(
        ValueError,
        match="Cannot calculate coherence",
    ):
        spectral.coherence(
            first,
            second,
            segmentlength=seglen,
            downsample=False,
        )

    # but that accepting downsampling gives you the same result as
    # doing nothing (but doesn't emit a warning)
    with warnings.catch_warnings():
        warnings.simplefilter("error", UserWarning)
        coh2 = spectral.coherence(
            first,
            second,
            segmentlength=seglen,
            downsample=True,
        )

    assert all(np.array(coh1.data) == np.array(coh2.data))



# Copyright (c) 2018-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwpy.plot.segments`."""

import numpy
import pytest
from matplotlib import rcParams
from matplotlib.collections import PatchCollection
from matplotlib.colors import ColorConverter

from ...segments import (
    DataQualityDict,
    DataQualityFlag,
    Segment,
    SegmentList,
    SegmentListDict,
)
from ...time import to_gps
from .. import SegmentAxes
from ..segments import SegmentRectangle
from .test_axes import TestAxes as _TestAxes

# extract color cycle
COLOR_CONVERTER = ColorConverter()
COLOR_CYCLE = rcParams["axes.prop_cycle"].by_key()["color"]
COLOR0 = COLOR_CONVERTER.to_rgba(COLOR_CYCLE[0])


class TestSegmentAxes(_TestAxes):
    """Tests for `SegmentAxes`."""

    AXES_CLASS = SegmentAxes

    @staticmethod
    @pytest.fixture
    def segments():
        """Create a `SegmentList` for use in tests."""
        return SegmentList([Segment(0, 3), Segment(6, 7)])

    @staticmethod
    @pytest.fixture
    def flag():
        """Create a `DataQualityFlag` for use in tests."""
        known = SegmentList([
            Segment(0, 3),
            Segment(6, 7),
        ])
        active = SegmentList([
            Segment(1, 2),
            Segment(3, 4),
            Segment(5, 7),
        ])
        return DataQualityFlag(
            name="Test segments",
            known=known,
            active=active,
        )

    def test_plot_flag(self, ax, flag):
        """Test `SegmentAxes.plot_flag`."""
        c = ax.plot_flag(flag)
        assert c.get_label() == flag.texname
        assert len(ax.collections) == 2
        assert ax.collections[0] is c

        flag.isgood = False
        c = ax.plot_flag(flag)
        assert tuple(c.get_facecolors()[0]) == (1., 0., 0., 1.)

        c = ax.plot_flag(flag, known={"facecolor": "black"})
        c = ax.plot_flag(flag, known="fancy")

    def test_plot_dict(self, ax, flag):
        """Test `SegmentAxes.plot_dict`."""
        dqd = DataQualityDict()
        dqd["a"] = flag
        dqd["b"] = flag

        colls = ax.plot_dict(dqd)
        assert len(colls) == len(dqd)
        assert all(isinstance(c, PatchCollection) for c in colls)
        assert colls[0].get_label() == "a"
        assert colls[1].get_label() == "b"

        colls = ax.plot_dict(dqd, label="name")
        assert colls[0].get_label() == "Test segments"
        colls = ax.plot_dict(dqd, label="anything")
        assert colls[0].get_label() == "anything"

    def test_plot_segmentlist(self, ax, segments):
        """Test `SegmentAxes.plot_segmentlist`."""
        coll = ax.plot_segmentlist(segments)
        assert isinstance(coll, PatchCollection)
        assert numpy.isclose(ax.dataLim.x0, 0.)
        assert numpy.isclose(ax.dataLim.x1, 7.)
        assert len(coll.get_paths()) == len(segments)
        assert ax.get_epoch() == segments[0][0]

    def test_plot_segmentlist_y(self, ax, segments):
        """Test `SegmentAxes.plot_segmentlist(..., y=N)`."""
        p = ax.plot_segmentlist(
            segments,
            valign="bottom",
        ).get_paths()[0].get_extents()
        assert p.y0 == 0.
        p = ax.plot_segmentlist(
            segments,
            valign="bottom",
            y=8,
        ).get_paths()[0].get_extents()
        assert p.y0 == 8.

    def test_plot_segmentlist_kwargs(self, ax, segments):
        """Test `SegmentAxes.plot_segmentlist` works with kwargs."""
        coll = ax.plot_segmentlist(
            segments,
            label="My segments",
            rasterized=True,
        )
        assert coll.get_label() == "My segments"
        assert coll.get_rasterized() is True

    def test_plot_segmentlist_collection_false(self, ax, segments):
        """Test `SegmentAxes.plot_segmentlist(..., collection=False)`."""
        coll = ax.plot_segmentlist(
            segments,
            collection=False,
            label="test",
        )
        assert isinstance(coll, list)
        assert not isinstance(coll, PatchCollection)
        assert coll[0].get_label() == "test"
        assert coll[1].get_label() == ""
        assert len(ax.patches) == len(segments)

    def test_plot_segmentlist_empty(self, ax):
        """Test that `SegmentAxes.plot_segmentlist` handles an empty segmentlist."""
        c = ax.plot_segmentlist(SegmentList())
        assert isinstance(c, PatchCollection)

    def test_plot_segmentlistdict(self, ax, segments):
        """Test `SegmentAxes.plot_segmentlistdict`."""
        sld = SegmentListDict()
        sld["TEST"] = segments
        ax.plot(sld)

    def test_plot(self, ax, segments, flag):  # type: ignore[override]
        """Test `SegmentAxes.plot`."""
        dqd = DataQualityDict(a=flag)
        ax.plot(segments)
        ax.plot(flag)
        ax.plot(dqd)
        ax.plot(flag, segments, dqd)

    def test_insetlabels(self, ax, segments):
        """Test `SegmentAxes.insetlabels` property."""
        ax.plot(segments)
        ax.set_insetlabels(True)

    def test_fmt_data(self, ax):
        """Test `SegmentAxes.format_xdata`."""
        # just check that the LIGOTimeGPS repr is in place
        value = 1234567890.123
        assert ax.format_xdata(value) == str(to_gps(value))

    # -- disable tests from upstream

    def test_imshow(self):  # type: ignore[override]
        """Test `SegmentAxes.imshow`."""
        pytest.skip(f"not implemented for {type(self).__name__}")


def test_segmentrectangle():
    """Test `SegmentRectangle`."""
    patch = SegmentRectangle((1.1, 2.4), 10)
    assert patch.get_xy(), (1.1, 9.6)
    assert numpy.isclose(patch.get_height(), 0.8)
    assert numpy.isclose(patch.get_width(), 1.3)
    assert patch.get_facecolor() == COLOR0


def test_segmentrectangle_kwargs():
    """Test `SegmentRectangle` handling of kwargs."""
    patch = SegmentRectangle(
        (1.1, 2.4),
        10,
        facecolor="red",
    )
    assert patch.get_facecolor() == COLOR_CONVERTER.to_rgba("red")


@pytest.mark.parametrize(("valign", "y"), [
    pytest.param("top", 9.2, id="top"),
    pytest.param("bottom", 10., id="bottom"),
])
def test_segmentrectangle_valign(valign, y):
    """Test `SegmentRectangle` handling of `valign`."""
    patch = SegmentRectangle((1.1, 2.4), 10, valign=valign)
    assert patch.get_xy() == (1.1, y)


def test_segmentrectangle_valign_error():
    """Test handling of invalid `valign` with `SegmentRectangle`."""
    with pytest.raises(
        ValueError,
        match="valign must be one of 'top', 'center', or 'bottom'",
    ):
        SegmentRectangle((0, 1), 0, valign="blah")  # type: ignore[arg-type]



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""
.. sectionauthor:: Duncan Macleod <duncan.macleod@ligo.org>
.. currentmodule:: gwpy.timeseries

Plotting a normalised `~gwpy.spectrogram.Spectrogram`
#####################################################

The :ref:`sphx_glr_examples_spectrogram_plot.py` example showed how to
generate and display a `~gwpy.spectrogram.Spectrogram` of the LIGO-Hanford
strain data around the |GW150914|_ event.

However, because of the shape of the LIGO sensitivity curve, picking out
features in the most sensitive frequency band (a few hundred Hertz) is
very hard.

We can normalise our `~gwpy.spectrogram.Spectrogram` to highligh those
features.
"""

# %%
# Again, we import the `TimeSeries` and call
# :meth:`TimeSeries.get` the download the strain
# data for the LIGO-Hanford interferometer
from gwpy.timeseries import TimeSeries
data = TimeSeries.get(
    "H1",
    "Sep 14 2015 09:45",
    "Sep 14 2015 09:55",
)

# %%
# Next, we can calculate a `~gwpy.spectrogram.Spectrogram` using the
# :meth:`spectrogram` method of the `TimeSeries` over a 2-second stride
# with a 1-second FFT and # .5-second overlap (50%):
specgram = data.spectrogram(2, fftlength=1, overlap=.5) ** (1/2.)

# %%
# and can normalise it against the overall median ASD by calling the
# :meth:`~gwpy.spectrogram.Spectrogram.ratio` method:

normalised = specgram.ratio("median")

# %%
# Finally, we can make a plot using the
# :meth:`~gwpy.spectrogram.Spectrogram.plot` method
plot = normalised.plot(norm="log", vmin=.1, vmax=10, cmap="Spectral_r")
ax = plot.gca()
ax.set_yscale("log")
ax.set_ylim(10, 2000)
ax.colorbar(label="Relative amplitude")
plot.show()

# %%
# Even with a normalised spectrogram, the resolution is such that a signal
# as short as that of GW150914 is impossible to see.
# See :ref:`sphx_glr_examples_spectrogram_spectrogram2.py` for an example that
# uses a high-resolution spectrogram method to zoom in around the exact time of
# the signal.



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.io.kerberos`."""

# ruff: noqa: S106

import os
import subprocess
from unittest import mock

import pytest

from .. import kerberos as io_kerberos

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def requires_gssapi():
    """Mark a function as requiring the `gssapi` module.

    This needs special handling because `import gssapi` on Windows
    can raise an `OSError` if a Kerberos implementation can't be
    loaded.
    """
    return pytest.mark.requires(
        "gssapi",
        exc_type=(ImportError, OSError),
    )


@pytest.fixture(autouse=True)
def mock_krb5_env():
    """Fixture to ensure Kerberos environment variables are not set."""
    with mock.patch.dict(os.environ):
        for key in (
            "KRB5_KTNAME",
            "KRB5CCNAME",
        ):
            os.environ.pop(key, None)
        yield


def kerberos_name(name):
    """Return a `gssapi.Name` for a Kerberos principal."""
    import gssapi
    return gssapi.Name(
        base=name,
        name_type=gssapi.NameType.kerberos_principal,
    )


@mock.patch.dict("sys.modules", {"gssapi": None})
def test_kinit_no_gssapi():
    """Test that we get an augmented error message if ``gssapi`` is missing."""
    with pytest.raises(
        ImportError,
        match="cannot generate Kerberos credentials without python-gssapi",
    ):
        io_kerberos.kinit()


@requires_gssapi()
@mock.patch("sys.stdout.isatty", return_value=True)
@mock.patch("gwpy.io.kerberos.input", return_value="rainer.weiss")
@mock.patch("getpass.getpass", return_value="test")
@mock.patch("gssapi.raw.acquire_cred_with_password")
@mock.patch("gssapi.Credentials")
def test_kinit_up(mock_creds, acquire, getpass, input_, mock_isatty):
    """Test `gwpy.io.kerberos.kinit` with username and password given.

    Note: without a real credential to use, we can't do much more than check
    that the function under test passes the right arguments to the GSSAPI
    library.
    """
    acquire.return_value = mock.MagicMock()

    # basic call should prompt for username and password
    io_kerberos.kinit(realm="LIGO.ORG")
    input_.assert_called_with(
        "Kerberos principal (user@LIGO.ORG): ",
    )
    getpass.assert_called_with(
        prompt="Password for rainer.weiss@LIGO.ORG: ",
    )

    # and then use the results in gssapi calls
    acquire.assert_called_with(
        kerberos_name("rainer.weiss@LIGO.ORG"),
        b"test",
        lifetime=None,
        usage="initiate",
    )


@requires_gssapi()
@mock.patch("gwpy.io.kerberos.input")
@mock.patch("getpass.getpass")
@mock.patch("gssapi.raw.acquire_cred_with_password")
@mock.patch("gssapi.Credentials")
def test_kinit_up_kwargs(mock_creds, acquire, getpass, input_):
    """Test `gwpy.io.kerberos.kinit` with username and password given.

    Note: without a real credential to use, we can't do much more than check
    that the function under test passes the right arguments to the GSSAPI
    library.
    """
    io_kerberos.kinit(
        username="albert.einstein@EXAMPLE.COM",
        password="test",
    )
    input_.assert_not_called()
    getpass.assert_not_called()
    acquire.assert_called_with(
        kerberos_name("albert.einstein@EXAMPLE.COM"),
        b"test",
        lifetime=None,
        usage="initiate",
    )


@requires_gssapi()
def test_kinit_keytab_dne(tmp_path):
    """Test `gwpy.io.kerberos.kinit` with a non-existent keytab."""
    keytab = tmp_path / "keytab"  # does not exist

    # check that missing keytab goes to password auth
    with pytest.raises(
        io_kerberos.KerberosError,
        match=rf"\AKerberos keytab .*{keytab.name}' is invalid",
    ):
        io_kerberos.kinit(
            username="test",
            password="passwd",
            keytab=keytab,
        )


@requires_gssapi()
@mock.patch.dict("os.environ")
@mock.patch("gssapi.Credentials")
@mock.patch("gwpy.io.kerberos._keytab_principal", return_value="rainer.weiss@LIGO.ORG")
def test_kinit_keytab(mock_keytab_principal, mock_creds, tmp_path):
    """Test `gwpy.io.kerberos.kinit` can handle keytabs properly."""
    keytab = tmp_path / "keytab"
    keytab.touch()
    ccache = tmp_path / "ccache"

    # test keytab kwarg
    io_kerberos.kinit(
        keytab=keytab,
        ccache=ccache,
        lifetime=1000,
    )
    mock_creds.assert_called_once_with(
        name=kerberos_name("rainer.weiss@LIGO.ORG"),
        store={
            "client_keytab": str(keytab),
            "ccache": str(ccache),
        },
        usage="initiate",
        lifetime=1000,
    )

    # pass keytab via environment
    mock_creds.reset()
    os.environ["KRB5_KTNAME"] = str(keytab)
    io_kerberos.kinit()
    mock_creds.assert_called_with(
        name=kerberos_name("rainer.weiss@LIGO.ORG"),
        store={
            "client_keytab": str(keytab),
        },
        usage="initiate",
        lifetime=None,
    )


@requires_gssapi()
@mock.patch("sys.stdout.isatty", return_value=False)
@mock.patch.dict("os.environ", clear=True)
def test_kinit_notty(mock_isatty):
    """Test `gwpy.io.kerberos.kinit` raises an error if it can't prompt."""
    with pytest.raises(io_kerberos.KerberosError):
        io_kerberos.kinit()


@requires_gssapi()
@mock.patch("gwpy.io.kerberos._acquire_password")
def test_kinit_error(mock_acquire_password):
    """Test that `gwpy.io.kerberos.kinit` propagates `GSSError`s appropriately."""
    import gssapi
    mock_acquire_password.side_effect = gssapi.exceptions.GSSError(0, 0)

    with pytest.raises(
        io_kerberos.KerberosError,
        match=(
            r"\Afailed to generate Kerberos TGT for test@EXAMPLE.COM"
        ),
    ):
        io_kerberos.kinit(
            username="test@EXAMPLE.COM",
            password="test",
        )


# -- deprecated

@requires_gssapi()
@mock.patch("gwpy.io.kerberos._acquire_password")
def test_kinit_krb5ccname(mock_acquire_password):
    """Test that the ``krb5ccname`` keyword emits a deprecation warning."""
    with pytest.warns(
        DeprecationWarning,
        match=(
            "The `krb5ccname` keyword for gwpy.io.kerberos.kinit was renamed"
        ),
    ):
        io_kerberos.kinit(
            username="test@EXAMPLE.COM",
            password="test",
            krb5ccname="cache",
        )


# mocked klist output
KLIST = b"""Keytab name: FILE:/test.keytab
KVNO Principal
---- -------------------------------
   1 albert.einstein@LIGO.ORG
   2 ronald.drever@LIGO.ORG
   2 ronald.drever@LIGO.ORG"""


@mock.patch("subprocess.check_output", return_value=KLIST)
def test_parse_keytab(mock_check_output):
    """Test `gwpy.io.kerberos.parse_keytab."""
    # assert principals get extracted correctly
    with pytest.deprecated_call():
        principals = io_kerberos.parse_keytab("test.keytab")
    assert principals == [("albert.einstein", "LIGO.ORG", 1),
                          ("ronald.drever", "LIGO.ORG", 2)]


@mock.patch("subprocess.check_output")
@pytest.mark.parametrize(("se", "match"), [
    (OSError, "^failed to locate klist, cannot read keytab$"),
    (
        subprocess.CalledProcessError(1, "error"),
        "cannot read keytab 'test.keytab'",
    ),
])
def test_parse_keytab_oserror(mock_check_output, se, match):
    """Test `gwpy.io.kerberos.parse_keytab` fails appropriately."""
    mock_check_output.side_effect = se
    with pytest.deprecated_call(), pytest.raises(
        io_kerberos.KerberosError,
        match=match,
    ):
        io_kerberos.parse_keytab("test.keytab")



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.time`."""

from datetime import (
    UTC,
    datetime,
)
from decimal import Decimal

import numpy
import pytest
from astropy.time import Time
from astropy.units import (
    Quantity,
    UnitConversionError,
)

from ... import time
from .. import LIGOTimeGPS

try:
    from glue.lal import LIGOTimeGPS as GlueGPS
except ImportError:
    GlueGPS = LIGOTimeGPS

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

GW150914 = LIGOTimeGPS(1126259462, 391000000)
GW150914_DT = datetime(2015, 9, 14, 9, 50, 45, 391000, tzinfo=UTC)
FREEZE = "2015-09-14 09:50:45.391"
NOW = 1126259462
TODAY = 1126224017
TOMORROW = 1126310417
YESTERDAY = 1126137617


@pytest.mark.freeze_time(FREEZE)
@pytest.mark.parametrize(("in_", "out"), [
    (1126259462, int(GW150914)),
    (1235635623.7500002, LIGOTimeGPS(1235635623, 750000200)),
    (LIGOTimeGPS(1126259462, 391000000), GW150914),
    ("0", 0),
    ("Jan 1 2017", 1167264018),
    ("Sep 14 2015 09:50:45.391", GW150914),
    ("Oct 30 2016 12:34 CST", 1161887657),
    ((2017, 1, 1), 1167264018),
    (datetime(2017, 1, 1, tzinfo=UTC), 1167264018),
    (Time(57754, format="mjd"), 1167264018),
    (Time(57754.0001, format="mjd"), LIGOTimeGPS(1167264026, 640000000)),
    (Quantity(1167264018, "s"), 1167264018),
    (Decimal("1126259462.391000000"), GW150914),
    pytest.param(
        GlueGPS(GW150914.gpsSeconds, GW150914.gpsNanoSeconds),
        GW150914,
        marks=pytest.mark.requires("glue"),
    ),
    (numpy.int32(NOW), NOW),  # fails with lal-6.18.0
    ("now", NOW),
    ("today", TODAY),
    ("tomorrow", TOMORROW),
    ("yesterday", YESTERDAY),
])
def test_to_gps(in_, out):
    """Test that :func:`to_gps` works."""
    assert time.to_gps(in_) == out


@pytest.mark.parametrize(("in_", "err"), [
    (Quantity(1, "m"), UnitConversionError),
    ("random string", ValueError),
])
def test_to_gps_error(in_, err):
    """Test that :func:`gwpy.time.to_gps` errors when it should."""
    with pytest.raises(err):
        time.to_gps(in_)


@pytest.mark.parametrize(("in_", "out"), [
    (1167264018, datetime(2017, 1, 1, tzinfo=UTC)),
    ("1167264018", datetime(2017, 1, 1, tzinfo=UTC)),
    (1126259462.391, datetime(2015, 9, 14, 9, 50, 45, 391000, tzinfo=UTC)),
    ("1.13e9", datetime(2015, 10, 27, 16, 53, 3, tzinfo=UTC)),
    pytest.param(
        GlueGPS(GW150914.gpsSeconds, GW150914.gpsNanoSeconds),
        GW150914_DT,
        marks=pytest.mark.requires("glue"),
    ),
])
def test_from_gps(in_, out):
    """Test that :func:`gwpy.time.from_gps` works."""
    assert time.from_gps(in_) == out


@pytest.mark.parametrize(("in_", "err"), [
    ("test", ValueError),
    (1167264017, ValueError),  # gwpy/gwpy#1021
])
def test_from_gps_error(in_, err):
    """Test that :func:`gwpy.time.from_gps` errors when it should."""
    with pytest.raises(err):
        time.from_gps(in_)


@pytest.mark.freeze_time(FREEZE)
@pytest.mark.parametrize(("in_", "out"), [
    (float(GW150914), GW150914_DT),
    (GW150914, GW150914_DT),
    (GW150914_DT, GW150914),
    pytest.param(
        GlueGPS(float(GW150914)),
        GW150914_DT,
        marks=pytest.mark.requires("glue"),
    ),
    ("now", NOW),
    ("today", TODAY),
    ("tomorrow", TOMORROW),
    ("yesterday", YESTERDAY),
])
def test_tconvert(in_, out):
    """Test :func:`gwpy.time.tconvert`."""
    assert time.tconvert(in_) == out


@pytest.mark.parametrize("gpsmod", [
    "glue.lal",
    "lal",
    "ligotimegps",
])
def test_gps_types(gpsmod):
    """Test that the module's `LIGOTimeGPS` matches the protocol."""
    mod = pytest.importorskip(gpsmod)
    gps = mod.LIGOTimeGPS(123, 456000000)
    assert isinstance(gps, time.LIGOTimeGPSLike)
    assert gps.gpsSeconds == 123
    assert gps.gpsNanoSeconds == 456000000
    assert gps == 123.456



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Input/Output utilities for GW Cache files.

A cache file is a specially-formatted ASCII file that contains file paths
and associated metadata for those files, designed to make identifying
relevant data, and sieving large file lists, easier for the user.
"""

from __future__ import annotations

import contextlib
import os
import warnings
from collections.abc import Sequence
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    NamedTuple,
    cast,
)

from ..segments import (
    Segment,
    SegmentList,
)
from ..time import LIGOTimeGPS
from .utils import (
    Readable,
    file_path,
    with_open,
)

if TYPE_CHECKING:
    from collections.abc import (
        Callable,
        Iterable,
        Iterator,
    )
    from typing import (
        IO,
        Any,
        Literal,
        TypeVar,
    )

    from lal.utils import CacheEntry

    from ..time import LIGOTimeGPSLike
    from .utils import (
        FileSystemPath,
        NamedReadable,
    )

    CacheEntryType = TypeVar("CacheEntryType", bound=FileSystemPath | CacheEntry)
    GpsParser = (
        Callable[[str], LIGOTimeGPSLike | float]
        | type[LIGOTimeGPS | LIGOTimeGPSLike | float]
    )

try:
    from lal.utils import CacheEntry
except ImportError:
    HAS_CACHEENTRY = False
else:
    HAS_CACHEENTRY = True

try:
    from glue.lal import Cache
except ImportError:
    HAS_CACHE = False
else:
    HAS_CACHE = True

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

__all__ = [
    "cache_segments",
    "file_segment",
    "filename_metadata",
    "find_contiguous",
    "flatten",
    "is_cache",
    "is_cache_entry",
    "read_cache",
    "read_cache_entry",
    "sieve",
    "write_cache",
]


@contextlib.contextmanager
def _silence_lal_debug_warnings() -> Iterator[None]:
    """Temporarily silence debug warnings from LAL."""
    try:
        import lal
    except ImportError:
        yield
        return

    debuglevel = lal.GetDebugLevel()
    # silence LAL debug warnings
    lal.ClobberDebugLevel(0)

    try:
        yield
    finally:
        # reset
        lal.ClobberDebugLevel(debuglevel)


def _preformat_entry(
    entry: NamedReadable | bytes | CacheEntry,
) -> tuple[str, str, str, Segment]:
    path = file_path(entry)
    obs, tag, seg = filename_metadata(path)
    start, end = seg
    if start.is_integer():
        start = int(start)
    if end.is_integer():
        end = int(end)
    return path, obs, tag, type(seg)(start, end)


def _format_entry_lal(entry: NamedReadable | bytes | CacheEntry) -> str:
    path, obs, tag, seg = _preformat_entry(entry)
    return f"{obs} {tag} {seg[0]} {abs(seg)} {path}"


def _parse_entry_lal(
    line: list[str],
    gpstype: GpsParser = LIGOTimeGPS,
) -> _CacheEntry:
    obs, desc, start, dur, path = line
    gpsstart = gpstype(start)
    end = gpsstart + float(dur)
    return _CacheEntry(obs, desc, Segment(gpsstart, end), path)


def _format_entry_ffl(entry: NamedReadable | bytes | CacheEntry) -> str:
    path, _, _, seg = _preformat_entry(entry)
    return f"{path} {seg[0]} {abs(seg)} 0 0"


def _parse_entry_ffl(
    line: list[str],
    gpstype: GpsParser = LIGOTimeGPS,
) -> _CacheEntry:
    path, start, dur, _, _ = line
    with _silence_lal_debug_warnings():
        gpsstart = gpstype(start)
    gpsend = gpsstart + float(dur)
    try:
        observatory, description = os.path.basename(path).split("-", 2)[:2]
    except ValueError:
        return _CacheEntry(None, None, Segment(gpsstart, gpsend), path)
    return _CacheEntry(observatory, description, Segment(gpsstart, gpsend), path)


class _CacheEntry(NamedTuple):
    """Quick version of lal.utils.CacheEntry for internal purposes only.

    Just to allow metadata handling for files that don't follow :dcc:`LIGO-T010150`.
    """

    observatory: str | None
    description: str | None
    segment: Segment
    path: str

    def __str__(self) -> str:
        return self.path

    def __fspath__(self) -> str:
        """Return the path component of the URL."""
        path = self.path
        if "://" in path:
            scheme, path = path.split("://", 1)
        else:
            scheme = None
        if scheme and scheme != "file":
            msg = (
                f"cannot use {type(self).__name__} as path-like object with "
                f"scheme='{scheme}'"
            )
            raise ValueError(msg)
        return path

    @classmethod
    def parse(
        cls,
        line: str | bytes,
        gpstype: GpsParser = LIGOTimeGPS,
    ) -> _CacheEntry:

        # format line string
        if isinstance(line, bytes):
            line = line.decode("utf-8")
        parts = line.strip().split()

        # if single entry, parse filename
        if len(parts) == 1:
            path = parts[0]
            return cls(*filename_metadata(path), path)

        try:
            return _parse_entry_ffl(parts, gpstype=gpstype)
        except (RuntimeError, TypeError, ValueError) as exc:
            try:
                return _parse_entry_lal(parts, gpstype=gpstype)
            except ValueError:
                pass
            exc.args = (
                f"Cannot identify format for cache entry {line.strip()!r}",
            )
            raise


# -- cache I/O -----------------------

def _iter_cache(
    cachefile: Iterable[str],
    gpstype: GpsParser = LIGOTimeGPS,
) -> Iterator[_CacheEntry]:
    """Yield a `_CacheEntry` for each line in the file.

    This method supports reading LAL- and (nested) FFL-format cache files.
    """
    try:
        path = Path(cachefile.name).resolve()  # type: ignore[attr-defined]
    except AttributeError:
        path = None
    for line in cachefile:
        try:
            yield _CacheEntry.parse(line, gpstype=gpstype)
        except ValueError:
            # virgo FFL format (seemingly) supports nested FFL files
            parts = line.split()
            if len(parts) == 3 and Path(parts[0]).resolve() != path:
                with Path(parts[0]).open() as cache2:
                    yield from _iter_cache(cache2, gpstype=gpstype)
            else:
                raise


@with_open
def read_cache(
    cachefile: Readable,
    coltype: GpsParser = LIGOTimeGPS,
    sort: Callable[[str], Any] | None = None,
    segment: Segment | None = None,
    *,
    strict: bool = False,
) -> list[str]:
    """Read a LAL- or FFL-format cache file as a list of file paths.

    Parameters
    ----------
    cachefile : `str`, `pathlib.Path`, `file`
        Input file or file path to read.

    coltype : `LIGOTimeGPS`, `int`, optional
        Type for GPS times.

    sort : `callable`, optional
        A callable key function by which to sort the output list of file paths

    segment : `gwpy.segments.Segment`, optional
        A GPS `[start, stop)` interval, if given only files overlapping this
        interval will be returned.

    strict : `bool`, optional
        If `False` warn about entries that don't follow the :dcc:`LIGO-T010150`
        standard, then skip them; if `True` all errors are raised as
        exceptions.

    Returns
    -------
    paths : `list` of `str`
        A list of file paths as read from the cache file.
    """
    # by now the decorator has opened the file for us
    cacheio = cast("Iterable[str]", cachefile)

    # read file
    cache = [x.path for x in _iter_cache(cacheio, gpstype=coltype)]

    # sieve and sort
    if segment:
        cache = sieve(cache, segment=segment, strict=strict)
    if sort:
        cache.sort(key=sort)

    # read simple paths
    return cache


def read_cache_entry(
    line: str | bytes,
    gpstype: GpsParser = LIGOTimeGPS,
) -> str:
    """Read a file path from a line in a cache file.

    Parameters
    ----------
    line : `str`, `bytes`
        Line of text to parse

    gpstype : `LIGOTimeGPS`, `int`, optional
        Type for GPS times.

    Returns
    -------
    path : `str`
       The file path.

    Raises
    ------
    ValueError
        if the line cannot be parsed successfully
    """
    return _CacheEntry.parse(line, gpstype=gpstype).path


@with_open(mode="w", pos=1)
def write_cache(
    cache: Iterable[FileSystemPath],
    fobj: IO,
    *,
    format: Literal["lal", "ffl"] | None = None,  # noqa: A002
) -> None:
    """Write a `list` of cache entries to a file.

    Parameters
    ----------
    cache : `list` of `str` or `os.PathLike`
        The list of path-like things to write.

    fobj : `file`, `str`, `pathlib.Path`
        The open file object, or file path to write to.

    format : `str`, optional
        The format to write to, one of

        - `None` : format each entry using `str`
        - ``'lal'`` : write a LAL-format cache
        - ``'ffl'`` : write an FFL-format cache
    """
    formatter: Callable[[FileSystemPath], str]
    if format is None:
        formatter = str
    elif format.lower() == "lal":
        formatter = _format_entry_lal
    elif format.lower() == "ffl":
        formatter = _format_entry_ffl
    else:
        msg = f"Unrecognised cache format {format!r}"
        raise ValueError(msg)

    # write file
    for line in map(formatter, cache):
        try:
            print(line, file=fobj)
        except TypeError:  # bytes-mode
            fobj.write(f"{line}\n".encode())


def is_cache(
    cache: Readable | Cache | Sequence[CacheEntry | FileSystemPath],
) -> bool:
    """Return `True` if ``cache`` is a readable cache file or object.

    Parameters
    ----------
    cache : `str`, `file`, `list`
        Object to detect as cache

    Returns
    -------
    iscache : `bool`
        `True` if the input object is a cache, or a file in LAL cache format,
        otherwise `False`
    """
    if isinstance(cache, Readable):
        try:
            return bool(len(read_cache(cache, coltype=float)))
        except Exception:  # noqa: BLE001
            # if parsing the file as a cache fails for _any reason_
            # presume it isn't a cache file
            return False
    if HAS_CACHE and isinstance(cache, Cache):
        return True
    return bool(
        isinstance(cache, Sequence)
        and cache
        and all(map(is_cache_entry, cache)),
    )


def is_cache_entry(path: CacheEntry | FileSystemPath) -> bool:
    """Return `True` if ``path`` can be represented as a cache entry.

    In practice this just tests whether the input is :dcc:`LIGO-T010150` compliant.

    Parameters
    ----------
    path : `str`, `lal.utils.CacheEntry`
        The input to test

    Returns
    -------
    isentry : `bool`
        `True` if ``path`` is an instance of `CacheEntry`, or can be parsed
        using :dcc:`LIGO-T010150`.
    """
    if HAS_CACHEENTRY and isinstance(path, CacheEntry):
        return True
    try:
        file_segment(path)
    except (ValueError, TypeError, AttributeError):
        return False
    return True


# -- cache manipulation --------------

def filename_metadata(filename: FileSystemPath) -> tuple[str, str, Segment]:
    """Return metadata parsed from a filename following :dcc:`LIGO-T010150`.

    This method is lenient with regards to integers in the GPS start time of
    the file, as opposed to `gwdatafind.utils.filename_metadata`, which is
    strict.

    Parameters
    ----------
    filename : `str`
        the path name of a file

    Returns
    -------
    obs : `str`
        the observatory metadata

    tag : `str`
        the file tag

    segment : `gwpy.segments.Segment`
        the GPS ``[float, float)`` interval for this file

    Notes
    -----
    :dcc:`LIGO-T010150` declares a file naming convention that includes
    documenting the GPS start integer and integer duration of a file,
    see that document for more details.

    Examples
    --------
    >>> from gwpy.io.cache import filename_metadata
    >>> filename_metadata("A-B-0-1.txt")
    ('A', 'B', Segment(0, 1))
    >>> filename_metadata("A-B-0.456-1.345.txt")
    ("A", "B", Segment(0.456, 1.801))
    """
    name = os.path.basename(filename)
    try:
        obs, desc, start, stub = name.split("-")
    except ValueError as exc:
        exc.args = (
            f"Failed to parse {name!r} as a LIGO-T010150-compatible filename",
        )
        raise
    gpsstart = float(start)
    dur: str = stub.rsplit(".", 1)[0]
    while True:  # recursively remove extension components
        try:
            fdur = float(dur)
        except ValueError:
            if "." not in dur:
                raise
            dur = dur.rsplit(".", 1)[0]
        else:
            break
    return obs, desc, Segment(gpsstart, gpsstart + fdur)


def file_segment(filename: CacheEntry | FileSystemPath) -> Segment:
    """Return the data segment for a filename following :dcc:`LIGO-T010150`.

    Parameters
    ----------
    filename : `str`, `~os.PathLike`, :class:`~lal.utils.CacheEntry`
        The path name of a file.

    Returns
    -------
    segment : `~gwpy.segments.Segment`
        the ``[start, stop)`` GPS segment covered by the given file

    Notes
    -----
    :dcc:`LIGO-T010150` declares a filenaming convention that includes
    documenting the GPS start integer and integer duration of a file,
    see that document for more details.
    """
    try:  # CacheEntry
        return Segment(filename.segment)  # type: ignore[union-attr]
    except AttributeError:  # file path (str)
        return filename_metadata(filename)[2]


def cache_segments(*caches: Iterable[CacheEntry | FileSystemPath]) -> SegmentList:
    """Return the segments of data covered by entries in the cache(s).

    Parameters
    ----------
    caches : `list`
        One or more lists of file paths (`str` or :class:`~lal.utils.CacheEntry`).

    Returns
    -------
    segments : `~gwpy.segments.SegmentList`
        A list of segments for when data should be available.
    """
    out = SegmentList()
    for cache in caches:
        out.extend(file_segment(e) for e in cache)
    return out.coalesce()


def flatten(
    *caches: list[CacheEntryType],
) -> list[CacheEntryType]:
    """Flatten a nested list of cache entries.

    Parameters
    ----------
    caches : `list`
        One or more lists of file paths (`str` or :class:`~lal.utils.CacheEntry`).

    Returns
    -------
    flat : `list`
        A flat `list` containing the unique set of entries across
        each input.
    """
    return list(dict.fromkeys(e for c in caches for e in c))


def find_contiguous(
    *caches: list[CacheEntryType],
) -> Iterator[list[CacheEntryType]]:
    """Separate one or more cache entry lists into time-contiguous sub-lists.

    Parameters
    ----------
    caches : `list`
        One or more lists of file paths (`str` or :class:`~lal.utils.CacheEntry`).

    Returns
    -------
    caches : `iter` of `list`
        An interable yielding each contiguous cache.
    """
    flat = flatten(*caches)
    for segment in cache_segments(flat):
        yield sieve(flat, segment=segment)


def sieve(
    cache: list[CacheEntryType],
    segment: Segment,
    *,
    strict: bool = True,
) -> list[CacheEntryType]:
    """Filter the cache to find those entries that overlap ``segment``.

    Parameters
    ----------
    cache : `list`
        Input list of file paths

    segment : `~gwpy.segments.Segment`
        The ``[start, stop)`` interval to match against.

    strict : `bool`, optional
        If `True` raise all exceptions, if `False` emit warnings when
        the file segment cannot be determined.

    Warns
    -----
    UserWarning
        For any files for which the file segment cannot be determined,
        if ``strict=False`` is given; these files are excluded from the
        sieved cache.
    """
    out = type(cache)()
    for e in cache:
        try:  # try and get the segment for this entry
            seg = file_segment(e)
        except ValueError as exc:
            # if running in 'strict' mode, raise all errors
            if strict:
                raise
            # otherwise warn and ignore
            warnings.warn(str(exc), stacklevel=2)
            continue
        if segment.intersects(seg):  # if overlaps, keep it
            out.append(e)
    return out



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read/write segment XML in LIGO_LW format into DataQualityFlags."""

from __future__ import annotations

import operator
from functools import reduce
from typing import TYPE_CHECKING

from ...io.ligolw import (
    build_content_handler,
    is_ligolw,
    read_ligolw,
    read_table,
    write_tables,
)
from ...io.registry import default_registry
from ...segments import (
    DataQualityDict,
    DataQualityFlag,
)

if TYPE_CHECKING:
    from typing import (
        Any,
        TypeAlias,
    )

    from igwn_ligolw.ligolw import (
        Document,
        PartialContentHandler,
    )

    from ...io.utils import (
        NamedReadable,
        Writable,
    )

    LigolwInput: TypeAlias = NamedReadable | list[NamedReadable]

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def segment_content_handler() -> PartialContentHandler:
    """Build a `~xml.sax.handlers.ContentHandler` to read segment XML tables."""
    from igwn_ligolw.ligolw import PartialLIGOLWContentHandler
    from igwn_ligolw.lsctables import (
        SegmentDefTable,
        SegmentSumTable,
        SegmentTable,
    )

    def _filter(name, attrs):
        return reduce(
            operator.or_,
            [table_.CheckProperties(name, attrs) for
             table_ in (SegmentTable, SegmentDefTable, SegmentSumTable)])

    return build_content_handler(PartialLIGOLWContentHandler, _filter)


# -- read ---------------------------------------------------------------------

def read_ligolw_dict(
    source: LigolwInput,
    names: list[str] | None = None,
    coalesce: bool = False,
    **kwargs,
) -> DataQualityDict:
    """Read segments for the given flag from the LIGO_LW XML file.

    Parameters
    ----------
    source : `file`, `str`, `list`
        One (or more) open files or file paths.

    names : `list`, `None`, optional
        List of names to read or `None` to read all into a single
        `DataQualityFlag`.

    coalesce : `bool`, optional
        If `True`, coalesce all parsed `DataQualityFlag` objects before
        returning, default: `False`

    kwargs
        Other keywords are passed to `DataQualityDict.from_ligolw_tables`.

    Returns
    -------
    flagdict : `DataQualityDict`
        A new `DataQualityDict` of `DataQualityFlag` entries with ``active``
        and ``known`` segments seeded from the XML tables in the given
        file ``fp``.
    """
    xmldoc = read_ligolw(source, contenthandler=segment_content_handler())
    segdef = read_table(xmldoc, "segment_definer")
    segsum = read_table(xmldoc, "segment_summary")
    seg = read_table(xmldoc, "segment")

    # parse tables
    out = DataQualityDict.from_ligolw_tables(
        segdef,
        segsum,
        seg,
        names=names,
        **kwargs,
    )

    # coalesce
    if coalesce:
        for flag in out:
            out[flag].coalesce()

    return out


def read_ligolw_flag(
    source: LigolwInput,
    name: str | None = None,
    **kwargs,
) -> DataQualityFlag:
    """Read a single `DataQualityFlag` from a LIGO_LW XML file."""
    return next(iter(read_ligolw_dict(
        source,
        names=[name] if name is not None else None,
        **kwargs,
    ).values()))


# -- write --------------------------------------------------------------------

def write_ligolw(
    flags: DataQualityFlag | DataQualityDict,
    target: Writable | Document,
    attrs: dict[str, Any] | None = None,
    **kwargs,
) -> None:
    """Write this `DataQualityFlag` to the given LIGO_LW Document.

    Parameters
    ----------
    flags : `DataQualityFlag`, `DataQualityDict`
        `gwpy.segments` object to write

    target : `str`, `file`, :class:`~igwn_ligolw.ligolw.Document`
        the file or document to write into

    attrs : `dict`, optional
        extra attributes to write into segment tables

    **kwargs
        keyword arguments to use when writing

    See Also
    --------
    gwpy.io.ligolw.write_ligolw_tables
        for details of acceptable keyword arguments
    """
    if isinstance(flags, DataQualityFlag):
        flags = DataQualityDict({flags.name: flags})
    return write_tables(
        target,
        flags.to_ligolw_tables(**attrs or {}),
        **kwargs,
    )


# -- register -----------------------------------------------------------------

# register methods for DataQualityDict
default_registry.register_reader("ligolw", DataQualityFlag, read_ligolw_flag)
default_registry.register_writer("ligolw", DataQualityFlag, write_ligolw)
default_registry.register_identifier("ligolw", DataQualityFlag, is_ligolw)

# register methods for DataQualityDict
default_registry.register_reader("ligolw", DataQualityDict, read_ligolw_dict)
default_registry.register_writer("ligolw", DataQualityDict, write_ligolw)
default_registry.register_identifier("ligolw", DataQualityDict, is_ligolw)



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""NDS2 data query routines for the TimeSeries."""

from __future__ import annotations

import importlib.util
import logging
import operator
from functools import reduce
from math import ceil
from typing import TYPE_CHECKING

from numpy import ones as numpy_ones

from ...detector import Channel
from ...io import nds2 as io_nds2
from ...segments import (
    Segment,
    SegmentList,
)
from ...time import to_gps
from ...utils.progress import progress_bar
from .. import (
    StateVector,
    StateVectorDict,
    TimeSeries,
    TimeSeriesDict,
)
from ..connect import _pad_series
from .losc import _any_gwosc_channels

if TYPE_CHECKING:
    from collections.abc import (
        Iterable,
        Sequence,
    )
    from typing import (
        SupportsFloat,
        TypeVar,
    )

    import nds2

    from ...time import SupportsToGps
    from .. import (
        TimeSeriesBase,
        TimeSeriesBaseDict,
    )

    _TChan = TypeVar("_TChan")
    _T = TypeVar("_T", bound=TimeSeriesBase)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

logger = logging.getLogger(__name__)


def log_nds2(
    connection: nds2.connection | str,
    message: str,
    *args: object,
    level: int = logging.DEBUG,
    **kwargs,
) -> None:
    """Emit a log message related to an open NDS2 connection."""
    if not isinstance(connection, str):
        connection = connection.get_host()
    message = f"[{connection}] {message}"
    logger.log(level, message, *args, **kwargs)


def _parse_nds_enum_dict_param(
    channels: Iterable[_TChan],
    key: str,
    value: dict | int | str | None,
) -> dict[_TChan, int]:
    """Parse an NDS2 Enum value into a dict keyed on ``channels``."""
    # parse input key
    enum: type[io_nds2._Nds2Enum]
    if key == "type":
        enum = io_nds2.Nds2ChannelType
        default = enum.any().value
    else:
        enum = io_nds2.Nds2DataType
        default = enum.any().value

    # set default
    if value is None:
        value = default

    # parse non-int enum representation
    if not isinstance(value, dict | int):
        value = enum.find(value).value

    # return dict of ints
    if isinstance(value, int):
        return dict.fromkeys(channels, value)

    # here we know ``value`` is a dict, so just fill in the blanks
    value = value.copy()
    for chan in channels:
        value.setdefault(chan, default)
    return value


def _set_parameter(
    connection: nds2.connection,
    parameter: str,
    value: str | int | bool,  # noqa: FBT001
) -> None:
    """Set a parameter for the connection, handling errors as warnings."""
    ref = f"{parameter}='{value}'"
    if connection.set_parameter(parameter, str(value)):
        log_nds2(connection, "Set %s", ref)
    else:
        log_nds2(connection, "Failed to set %s", ref, level=logging.WARNING)



def fetch_series(
    channel: str | Channel,
    start: SupportsToGps,
    end: SupportsToGps,
    *,
    type: str | int | dict | None = None,  # noqa: A002
    dtype: str | int | dict | None = None,
    allow_tape: bool | None = None,
    connection: nds2.connection | None = None,
    host: str | None = None,
    port: int | None = None,
    pad: float | None = None,
    scaled: bool | None = None,
    series_class: type[_T] = TimeSeries,
    verbose: bool | str = False,
) -> _T:
    """Fetch a single data series from NDS2.

    channel : `str`, `~gwpy.detector.Channel`
        The name (or representation) of the data channel to fetch.

    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
        GPS start time of required data,
        any input parseable by `~gwpy.time.to_gps` is fine

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS end time of required data, defaults to end of data found;
        any input parseable by `~gwpy.time.to_gps` is fine

    host : `str`, optional
        URL of NDS server to use, if blank will try any server
        (in a relatively sensible order) to get the data

        One of ``connection`` or ``host`` must be given.

    port : `int`, optional
        Port number for NDS server query, must be given with `host`.

    verify : `bool`, optional
        Check channels exist in database before asking for data.
        Default is `True`.

    verbose : `bool`, optional
        This argument is deprecated and will be removed in a future release.
        Use DEBUG-level logging instead, see :ref:`gwpy-logging`.

    connection : `nds2.connection`, optional
        Open NDS connection to use.
        Default is to open a new connection using ``host`` and ``port``
        arguments.

        One of ``connection`` or ``host`` must be given.

    pad : `float`, optional
        Float value to insert between gaps.
        Default behaviour is to raise an exception when any gaps are
        found.

    scaled : `bool`, optional
        Apply slope and bias calibration to ADC data, for non-ADC data
        this option has no effect.

    allow_tape : `bool`, optional
        Allow data access from slow tapes.
        If ``host`` or ``connection`` is given, the default is to do
        whatever the server default is, otherwise servers will be searched
        with ``allow_tape=False`` first, then ``allow_tape=True`` if that
        fails.

    type : `int`, `str`, optional
        NDS2 channel type integer or string name to match.
        Default is to search for any channel type.

    dtype : `numpy.dtype`, `str`, `type`, or `dict`, optional
        NDS2 data type to match.
        Default is to search for any data type.

    Returns
    -------
    data : `TimeSeries` or `StateVector`
        A new `TimeSeries` or `StateVector` fetched from NDS.
    """
    return fetch_dict(
        [channel],
        start,
        end,
        type=type,
        dtype=dtype,
        allow_tape=allow_tape,
        connection=connection,
        host=host,
        port=port,
        pad=pad,
        scaled=scaled,
        series_class=series_class,
        verbose=verbose,
    )[str(channel)]


def fetch_dict(
    channels: Sequence[str | Channel],
    start: SupportsToGps,
    end: SupportsToGps,
    *,
    type: str | int | dict | None = None,  # noqa: A002
    dtype: str | int | dict | None = None,
    allow_tape: bool | None = None,
    connection: nds2.connection | None = None,
    host: str | None = None,
    port: int | None = None,
    pad: float | None = None,
    scaled: bool | None = None,
    series_class: type[_T] = TimeSeries,
    verbose: bool | str = False,
) -> TimeSeriesBaseDict[_T]:
    """Fetch a dict of series data from NDS2.

    Parameters
    ----------
    channels : `list` of `str` or `Channel`
        List of channel names to fetch.

    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
        GPS start time of required data,
        any input parseable by `~gwpy.time.to_gps` is fine

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS end time of required data, defaults to end of data found;
        any input parseable by `~gwpy.time.to_gps` is fine

    host : `str`, optional
        URL of NDS server to use, if blank will try any server
        (in a relatively sensible order) to get the data

        One of ``connection`` or ``host`` must be given.

    port : `int`, optional
        Port number for NDS server query, must be given with `host`.

    verify : `bool`, optional
        Check channels exist in database before asking for data.
        Default is `True`.

    verbose : `bool`, optional
        This argument is deprecated and will be removed in a future release.
        Use DEBUG-level logging instead, see :ref:`gwpy-logging`.

    connection : `nds2.connection`, optional
        Open NDS connection to use.
        Default is to open a new connection using ``host`` and ``port``
        arguments.

        One of ``connection`` or ``host`` must be given.

    pad : `float`, optional
        Float value to insert between gaps.
        Default behaviour is to raise an exception when any gaps are
        found.

    scaled : `bool`, optional
        Apply slope and bias calibration to ADC data, for non-ADC data
        this option has no effect.

    allow_tape : `bool`, optional
        Allow data access from slow tapes.
        If ``host`` or ``connection`` is given, the default is to do
        whatever the server default is, otherwise servers will be searched
        with ``allow_tape=False`` first, then ``allow_tape=True`` if that
        fails.

    type : `int`, `str`, optional
        NDS2 channel type integer or string name to match.
        Default is to search for any channel type.

    dtype : `numpy.dtype`, `str`, `type`, or `dict`, optional
        NDS2 data type to match.
        Default is to search for any data type.

    Returns
    -------
    data : `TimeSeriesBaseDict`
        A new `TimeSeriesBaseDict` of (`str`, `TimeSeries`) pairs fetched
        from NDS.
    """
    # format GPS times
    gpsstart = to_gps(start)
    gpsend = to_gps(end)

    # format verbose progress
    desc: str
    if isinstance(verbose, str):
        desc = verbose
        verbose = True
    else:
        desc = "Downloading data"
        verbose = bool(verbose)

    # get a connection to an NDS2 server
    if connection:
        log_nds2(connection, "Using existing connection")
    else:
        if host is None:
            msg = (
                "either connection= or host= keyword arguments "
                "must be given to fetch data from NDS2"
            )
            raise ValueError(msg)
        connection = io_nds2.auth_connect(host, port)

    # set ALLOW_DATA_ON_TAPE
    if allow_tape is not None:
        _set_parameter(
            connection,
            "ALLOW_DATA_ON_TAPE",
            str(allow_tape),
        )

    ctype = _parse_nds_enum_dict_param(channels, "type", type)
    dtype = _parse_nds_enum_dict_param(channels, "dtype", dtype)

    # read using integers
    istart = int(gpsstart)
    iend = ceil(gpsend)

    # verify channels exist
    log_nds2(connection, "Checking channels list against database")
    utype = reduce(operator.or_, ctype.values())  # logical OR of types
    udtype = reduce(operator.or_, dtype.values())
    epoch = (istart, iend) if connection.get_protocol() > 1 else None
    ndschannels = io_nds2.find_channels(
        channels,
        connection=connection,
        epoch=epoch,
        type=utype,
        dtype=udtype,
        unique=True,
    )

    names = [Channel.from_nds2(c).ndsname for c in ndschannels]
    log_nds2(connection, "Channel check complete, %s names found", len(names))

    # handle minute trend timing
    if any(c.endswith("m-trend") for c in names) and (istart % 60 or iend % 60):
        log_nds2(
            connection,
            "Requested at least one minute trend, but "
            "start and stop GPS times are not multiples of "
            "60; times will be expanded outwards to compensate",
        )
        istart, iend = io_nds2.minute_trend_times(istart, iend)

    # get data availability
    span = SegmentList([Segment(istart, iend)])
    if pad is None:
        qsegs = span
        gap = "raise"
    elif connection.get_protocol() == 1:
        qsegs = span
        gap = "pad"
    else:
        log_nds2(connection, "Querying for data availability...")
        pad = float(pad)
        gap = "pad"
        qsegs = _get_data_segments(ndschannels, istart, iend, connection) & span
        log_nds2(
            connection,
            "Availability check complete, found %s viable segments of data "
            "with %.2f%% coverage",
            len(qsegs),
            abs(qsegs) / abs(span) * 100,
        )
        if span - qsegs:
            log_nds2(connection, "Gaps will be padded with %s", pad)

    # query for each segment
    out = series_class.DictClass()
    with progress_bar(
        total=float(abs(qsegs)),
        desc=desc,
        unit="s",
        disable=not bool(verbose),
    ) as bar:
        for seg in qsegs:
            total = 0.
            for buffers in connection.iterate(int(seg[0]), int(seg[1]), names):
                for buffer_, chan in zip(buffers, channels, strict=True):
                    series = series_class.from_nds2_buffer(
                        buffer_,
                        scaled=scaled,
                        copy=chan not in out,  # only copy if first buffer
                    )
                    out.append({chan: series}, pad=pad, gap=gap)
                new = buffer_.length / buffer_.channel.sample_rate
                total += new
                bar.update(new)
            # sometimes NDS2 returns no data at all
            if not total and gap != "pad":
                msg = f"no data received from {connection.get_host()} for {seg}"
                raise RuntimeError(msg)

    # finalise timeseries to make sure each channel has the correct limits
    # only if user asked to pad gaps
    if pad is not None:
        for chan, ndschan in zip(channels, ndschannels, strict=True):
            try:
                ts = out[chan]
            except KeyError:
                out[chan] = _create_series(
                    ndschan,
                    pad,
                    gpsstart,
                    gpsend,
                    series_class=series_class,
                )
            else:
                out[chan] = _pad_series(ts, pad, gpsstart, gpsend)

    # constrain to the non-integer GPS times we were actually given
    if gpsstart != istart or gpsend != iend:
        out.crop(start=gpsstart, end=gpsend)

    return out


def _create_series(
    ndschan: nds2.channel,
    value: float,
    start: SupportsFloat,
    end: SupportsFloat,
    series_class: type[TimeSeriesBase] = TimeSeries,
) -> TimeSeriesBase:
    """Create a timeseries to cover the specified [start, end) limits.

    To cover a gap in data returned from NDS.
    """
    channel = Channel.from_nds2(ndschan)
    nsamp = int((end - start) * channel.sample_rate.value)  # type: ignore[union-attr]
    return series_class(
        numpy_ones(nsamp) * value,
        t0=start,
        sample_rate=channel.sample_rate,
        unit=channel.unit,
        channel=channel)


def _get_data_segments(
    channels: list[nds2.channel],
    start: int,
    end: int,
    connection: nds2.connection,
) -> SegmentList:
    """Get available data segments for the given channels."""
    allsegs = io_nds2.get_availability(
        channels,
        start,
        end,
        connection=connection,
    )
    return allsegs.intersection(allsegs.keys())


# -- get registry --------------------


def identify_nds2_sources(
    origin: str,
    channels: str | Channel | Iterable[str | Channel],
    start: SupportsToGps | None,
    *args,  # noqa: ARG001
    connection: nds2.connection | None = None,
    host: str | None = None,
    port: int | None = None,
    allow_tape: bool | None = None,
    **kwargs,  # noqa: ARG001
) -> Iterable[dict[str, object]] | None:
    """Identify NDS2 sources for these arguments."""
    # NDS2 only works for 'get'
    if origin != "get":
        return None

    # We must be able to import the NDS2 client
    if importlib.util.find_spec("nds2") is None:
        return None

    # If the host is a GWDataFind NDS server, don't get involved
    if "datafind" in str(host):
        return []

    if connection is not None or host is not None:
        return [{
            "connection": connection,
            "host": host,
            "port": port,
            "allow_tape": allow_tape,
        }]

    channels = (channels,) if isinstance(channels, (str, Channel)) else channels

    # Just IFO names, not channels, most likely a GWOSC request
    if all(len(str(c)) <= 2 for c in channels):
        return []

    # Now iterate over possible NDS2 sources by combining known options
    # for host/port and allow_tape:

    sources: list[dict[str, object]] = []

    ifos = {Channel(channel).ifo for channel in channels}
    try:
        ifo = ifos.pop()
    except KeyError:  # ifos is empty
        ifo = None

    hostlist = io_nds2.host_resolution_order(ifo, epoch=start)

    if allow_tape is None:
        tapes = [False, True]
    else:
        tapes = [allow_tape]

    gwosc = _any_gwosc_channels(channels)

    for allow_tape_ in tapes:
        for host_, port_ in hostlist:
            # Deprioritise tape-allowed connections
            priority = 100 if allow_tape_ else 10
            if gwosc and "gwosc" in host_:
                # Prioritise the GWOSC server for GWOSC channels
                priority -= 1
            sources.append({
                "connection": None,
                "host": host_,
                "port": port_,
                "allow_tape": allow_tape_,
                "priority": priority,
            })

    return sources


for klass, fetch in (
    (TimeSeries, fetch_series),
    (StateVector, fetch_series),
    (TimeSeriesDict, fetch_dict),
    (StateVectorDict, fetch_dict),
):
    klass.get.registry.register_identifier(
        "nds2",
        klass,
        identify_nds2_sources,
    )
    klass.get.registry.register_reader(
        "nds2",
        klass,
        fetch,
    )



# Copyright (c) 2015-2020 Joseph Areeda
#               2020-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Base class for CLI (`gwpy-plot`) products."""

from __future__ import annotations

import abc
import logging
import os.path
import re
import time
import warnings
from functools import wraps
from typing import (
    TYPE_CHECKING,
    cast,
)

from astropy.time import Time
from astropy.units import Quantity
from matplotlib import rcParams, style

from ..plot.colors import GW_OBSERVATORY_COLORS
from ..plot.gps import GPS_SCALES, GPSTransform
from ..plot.tex import label_to_latex
from ..segments import DataQualityFlag
from ..signal import filter_design
from ..signal.window import recommended_overlap
from ..time import to_gps
from ..timeseries import (
    TimeSeries,
    TimeSeriesDict,
)
from ..timeseries.timeseries import DEFAULT_FFT_METHOD
from ..utils import unique

if TYPE_CHECKING:
    from argparse import (
        ArgumentParser,
        Namespace,
        _ArgumentGroup,
    )
    from collections.abc import Callable
    from logging import Logger
    from os import PathLike
    from typing import (
        BinaryIO,
        ClassVar,
        Literal,
        ParamSpec,
        TypeVar,
    )

    from astropy.units import UnitBase
    from astropy.units.typing import QuantityLike
    from matplotlib.legend import Legend

    from ..plot import (
        Axes,
        Plot,
    )
    from ..signal.filter_design import ZpkType

    P = ParamSpec("P")
    R = TypeVar("R")

    TimeSeriesType = TypeVar("TimeSeriesType", bound=TimeSeries)

__author__ = "Joseph Areeda <joseph.areeda@ligo.org>"

BAD_UNITS = {
    "*",
    "Counts.",
}

#: GPS value below which it's probably not a GPS
MIN_GPS = 1e8

logger = logging.getLogger(__name__)


# -- utilities -----------------------

def timer(func: Callable[P, R]) -> Callable[P, R]:
    """Decorate a function to time itself and log its duration after return."""
    name = func.__name__

    @wraps(func)
    def timed_func(*args: P.args, **kwargs: P.kwargs) -> R:
        inst = cast("CliProduct", args[0])
        _start = time.time()
        out = func(*args, **kwargs)
        inst.logger.debug("%s took %.1f sec", name, time.time() - _start)
        return out

    return timed_func


def to_float(unit: str | UnitBase) -> Callable:
    """Return a function to convert a qantity string to a float in the given unit.

    Examples
    --------
    >>> conv = to_float('Hz')
    >>> conv('4 mHz')
    >>> 0.004
    """

    def converter(x: QuantityLike) -> float:
        return Quantity(x, unit).value

    converter.__doc__ = f"Convert the input to a `float` in {unit}."
    return converter


to_hz = to_float("Hz")
to_s = to_float("s")


# -- base product class --------------

class CliProduct(metaclass=abc.ABCMeta):
    """Base class for all cli plot products.

    Parameters
    ----------
    args : `argparse.Namespace`
        Command-line arguments as parsed using
        :meth:`~argparse.ArgumentParser.parse_args`

    Notes
    -----
    This object has two main entry points,

    - `CliProduct.init_cli` - which adds arguments and argument groups to
       an `~argparse.ArgumentParser`
    - `CliProduct.run` - executes the arguments to product one or more plots

    So, the basic usage is follows::

    >>> import argparse
    >>> parser = argparse.ArgumentParser()
    >>> from gwpy.cli import CliProduct
    >>> CliProduct.init_cli(parser)
    >>> product = CliProduct(parser.parse_args())
    >>> product.run()

    The key methods for subclassing are

    - `CliProduct.action` - property defines 'name' for command-line subcommand
    - `CliProduct._finalize_arguments` - this is called from `__init__`
      to set defaults arguments for products that weren't set from the command
      line
    - `CliProduct.make_plot` - post-processes timeseries data and generates
      one figure
    """

    MIN_CHANNELS: ClassVar[int] = 1
    MIN_DATASETS: ClassVar[int] = 1
    MAX_DATASETS: ClassVar[int] = int(1e100)

    action: ClassVar[str | None] = None

    def __init__(
        self,
        args: Namespace,
        logger: Logger = logger,
    ) -> None:
        """Initialise this `CliProduct`."""
        #: input argument Namespace
        self.args = args

        #: Logger
        self.logger = logger

        # NB: finalizing may want to log if we're being verbose
        self._finalize_arguments(args)  # post-process args

        if args.style:  # apply custom styling
            style.use(args.style)

        #: the current figure object
        self.plot: Plot | None = None

        #: figure number
        self.plot_num = 0

        #: start times for data sets
        self.start_list = unique(map(
            int,
            (gps for gpsl in args.start for gps in gpsl),
        ))

        #: duration of each time segment
        self.duration = args.duration

        #: channels to load
        self.chan_list: list[str] = unique(c for clist in args.chan for c in clist)

        # use reduced as an alias for rds in channel name
        for idx in range(len(self.chan_list)):
            self.chan_list[idx] = self.chan_list[idx].replace("reduced", "rds")

        # total number of datasets that _should_ be acquired
        self.n_datasets = len(self.chan_list) * len(self.start_list)

        #: list of input data series (populated by get_data())
        self.timeseries: list[TimeSeries] = []

        #: dots-per-inch for figure
        self.dpi = args.dpi
        #: width and height in pixels
        self.width, self.height = map(float, self.args.geometry.split("x", 1))
        #: figure size in inches
        self.figsize = (self.width / self.dpi, self.height / self.dpi)
        #: Flag for data validation (like all zeroes)
        self.got_error = False

        # please leave this last
        self._validate_arguments()

    # -- abstract methods ------------

    @abc.abstractmethod
    def make_plot(self) -> Plot | None:
        """Generate the plot from time series and arguments.

        This method must be overridden by subclasses.

        Returns
        -------
        plot : `~gwpy.plot.Plot` or `None`
            The generated plot, or `None` if no plot could be made.
        """

    # -- properties ------------------

    @property
    def plot(self) -> Plot:
        """The current `~gwpy.plot.Plot` of this product.

        Raises a `RuntimeError` if the plot has not yet been created.
        """
        msg = "plot has not been created yet"
        try:
            plot = self._plot
        except AttributeError as exc:
            raise RuntimeError(msg) from exc
        if plot is None:
            raise RuntimeError(msg)
        return plot

    @plot.setter
    def plot(self, plot: Plot | None) -> None:
        """Set the current `~gwpy.plot.Plot` of this product."""
        self._plot = plot

    @property
    def axes(self) -> list[Axes]:
        """The list of `~matplotlib.axes.Axes` of this product's plot."""
        return cast("list[Axes]", self.plot.axes)

    @property
    def ax(self) -> Axes:
        """The current `~matplotlib.axes.Axes` of this product's plot."""
        return cast("Axes", self.plot.gca())

    @property
    def units(self) -> list[UnitBase]:
        """The (unique) list of data units for this product."""
        return unique(ts.unit for ts in self.timeseries)

    @property
    def usetex(self) -> bool:
        """Switch denoting whether LaTeX will be used or not."""
        return rcParams["text.usetex"]

    # -- utilities -------------------

    def log(
        self,
        level: int,
        msg: str,
        *args: object,
        **kwargs,
    ) -> None:
        """Print log message if verbosity is set high enough."""
        warnings.warn(
            "`CliProduct.log` is deprecated, use `CliProduct.logger` instead",
            DeprecationWarning,
            stacklevel=2,
        )
        # Convert verbosity level int to logging level
        lvl = max(3 - level, 0) * 10
        self.logger.log(lvl, msg, *args, **kwargs)

    # -- argument parsing ------------

    # each method below is a classmethod so that the command-line
    # for a product can be set up without having to create an instance
    # of the class

    @classmethod
    def init_cli(cls, parser: ArgumentParser) -> None:
        """Set up the argument list for this product."""
        cls.init_data_options(parser)
        cls.init_plot_options(parser)

    @classmethod
    def init_data_options(cls, parser: ArgumentParser) -> None:
        """Set up data input and signal processing options."""
        cls.arg_channels(parser)
        cls.arg_data(parser)
        cls.arg_signal(parser)

    @classmethod
    def init_plot_options(cls, parser: ArgumentParser) -> None:
        """Set up plotting options."""
        cls.arg_plot(parser)
        cls.arg_xaxis(parser)
        cls.arg_yaxis(parser)

    # -- data options

    @classmethod
    def arg_channels(cls, parser: ArgumentParser) -> _ArgumentGroup:
        """Add an `~argparse._ArgumentGroup` for channel options."""
        group = parser.add_argument_group(
            "Data options",
            "What data to load",
        )
        group.add_argument(
            "--chan",
            "--channel",
            "--ifo",
            type=str,
            nargs="+",
            action="append",
            required=True,
            help="channels to load (or IFO prefix for GWOSC data)",
        )
        group.add_argument(
            "--start",
            type=to_gps,
            nargs="+",
            action="append",
            required=True,
            help=("one or more starting times (integer GPS or date/time string)"),
        )
        group.add_argument(
            "--duration",
            type=to_s,
            default=10,
            help="Duration (seconds) [10]",
        )
        return group

    @classmethod
    def arg_data(cls, parser: ArgumentParser) -> _ArgumentGroup:
        """Add an `~argparse._ArgumentGroup` for data options."""
        group = parser.add_argument_group(
            "Data source options",
            "Where to get the data",
        )
        meg = group.add_mutually_exclusive_group()
        meg.add_argument(
            "-c",
            "--framecache",
            type=os.path.abspath,
            help=(
                "path to a file containing a list of paths from which to read data"
            ),
        )
        meg.add_argument(
            "-n",
            "--nds2-server",
            metavar="HOSTNAME",
            help="name of nds2 server to use, default is to try all of them",
        )
        meg.add_argument(
            "--frametype",
            help="GWF frametype to read from",
        )
        meg.add_argument(
            "--gwosc",
            action="store_true",
            default=False,
            help="get data from GWOSC",
        )
        return group

    @classmethod
    def arg_signal(cls, parser: ArgumentParser) -> _ArgumentGroup:
        """Add an `~argparse._ArgumentGroup` for signal-processing options."""
        group = parser.add_argument_group(
            "Signal processing options",
            "What to do with the data before plotting",
        )
        group.add_argument(
            "--highpass",
            type=to_hz,
            help="Frequency for highpass filter",
        )
        group.add_argument(
            "--lowpass",
            type=to_hz,
            help="Frequency for lowpass filter",
        )
        group.add_argument(
            "--notch",
            type=to_hz,
            nargs="*",
            help="Frequency for notch (can give multiple)",
        )
        return group

    # -- plot options

    @classmethod
    def arg_plot(cls, parser: ArgumentParser) -> _ArgumentGroup:
        """Add an `~argparse._ArgumentGroup` for basic plot options."""
        group = parser.add_argument_group("Plot options")
        group.add_argument(
            "-g",
            "--geometry",
            default="1200x600",
            metavar="WxH",
            help="size of resulting image",
        )
        group.add_argument(
            "--dpi",
            type=int,
            default=rcParams["figure.dpi"],
            help="dots-per-inch for figure",
        )
        group.add_argument(
            "--interactive",
            action="store_true",
            help=(
                "presents display in an interactive window, see "
                "https://matplotlib.org/stable/users/explain/"
                "interactive.html#default-ui for details"
            ),
        )
        group.add_argument(
            "--title",
            action="store",
            help="Set title (below suptitle, defaults to parameter summary",
        )
        group.add_argument(
            "--suptitle",
            help="topmost title line (larger than the others)",
        )
        group.add_argument(
            "--out",
            default="gwpy.png",
            help=(
                "output filename (extension determines format: "
                "png, pdf, svg are available)"
            ),
        )

        # legends match input files in position are displayed if specified.
        group.add_argument(
            "--legend",
            nargs="+",
            action="append",
            default=[],
            help="strings to match data files",
        )
        group.add_argument(
            "--nolegend",
            action="store_true",
            help="do not display legend",
        )
        group.add_argument(
            "--nogrid",
            action="store_true",
            help="do not display grid lines",
        )

        # allow custom styling with a style file
        group.add_argument(
            "--style",
            metavar="FILE",
            help=(
                "path to custom matplotlib style sheet, see "
                "http://matplotlib.org/users/style_sheets.html#style-sheets "
                "for details of how to write one"
            ),
        )
        return group

    @classmethod
    def arg_xaxis(cls, parser: ArgumentParser) -> _ArgumentGroup:
        """Add an `~argparse._ArgumentGroup` for X-axis options."""
        return cls._arg_axis("x", parser)

    @classmethod
    def arg_yaxis(cls, parser: ArgumentParser) -> _ArgumentGroup:
        """Add an `~argparse._ArgumentGroup` for Y-axis options."""
        return cls._arg_axis("y", parser)

    @classmethod
    def _arg_axis(
        cls,
        axis: str,
        parser: ArgumentParser,
        **defaults,
    ) -> _ArgumentGroup:
        name = f"{axis.title()} axis"
        group = parser.add_argument_group(f"{name} options")

        # label
        group.add_argument(
            f"--{axis}label",
            default=defaults.get("label"),
            dest=f"{axis}label",
            help=f"{name} label",
        )

        # min and max
        for extrema in ("min", "max"):
            opt = axis + extrema
            group.add_argument(
                f"--{opt}",
                type=float,
                default=defaults.get(extrema),
                dest=opt,
                help=f"{extrema} value for {name}",
            )

        # scale
        scaleg = group.add_mutually_exclusive_group()
        scaleg.add_argument(
            f"--{axis}scale",
            type=str,
            default=defaults.get("scale"),
            dest=f"{axis}scale",
            help=f"scale for {name}",
        )
        if defaults.get("scale") == "log":
            scaleg.add_argument(
                f"--nolog{axis}",
                action="store_const",
                dest=f"{axis}scale",
                const=None,
                default="log",
                help=f"use linear {name}",
            )
        else:
            scaleg.add_argument(
                f"--log{axis}",
                action="store_const",
                dest=f"{axis}scale",
                const="log",
                default=None,
                help=f"use logarithmic {name}",
            )
        return group

    def _finalize_arguments(self, args: Namespace) -> None:
        """Sanity-check and set defaults for arguments."""
        # this method is called by __init__ (after command-line arguments
        # have been parsed)

        if args.out is None:
            args.out = "gwpy.png"

    def _validate_arguments(self) -> None:
        """Sanity check arguments and raise errors if required."""
        # validate number of data sets requested
        if len(self.chan_list) < self.MIN_CHANNELS:
            msg = f"this product requires at least {self.MIN_CHANNELS} channels"
            raise ValueError(msg)
        if self.n_datasets < self.MIN_DATASETS:
            msg = (
                f"{self.MIN_DATASETS} are required for this plot but only "
                f"{self.n_datasets} are supplied"
            )
            raise ValueError(msg)
        if self.n_datasets > self.MAX_DATASETS:
            msg = (
                f"A maximum of {self.MAX_DATASETS} datasets allowed for this "
                f"plot but {self.n_datasets} specified"
            )
            raise ValueError(msg)

    # -- data transfer ---------------

    @timer
    def get_data(self) -> None:
        """Get all the data.

        This method populates the `timeseries` list attribute
        """
        args = self.args

        # determine how we're supposed get our data
        if args.gwosc:
            source = "gwosc"
        elif args.framecache is not None:
            source = "cache"
        else:
            source = "nds2"

        # Get the data from NDS or Frames
        for start in self.start_list:
            end = start + self.duration
            if source.lower() == "gwosc":
                tsd = TimeSeriesDict()
                for chan in self.chan_list:
                    ifo = chan.split(":", 1)[0]
                    data = TimeSeries.fetch_open_data(
                        ifo,
                        start,
                        end,
                    )
                    data.name = f"{ifo}:{data.name}"
                    tsd.append({data.name: data})
            elif source.lower() == "nds2":
                tsd = TimeSeriesDict.get(
                    self.chan_list,
                    start,
                    end,
                    host=args.nds2_server,
                    frametype=args.frametype,
                )
            else:
                tsd = TimeSeriesDict.read(
                    args.framecache,
                    self.chan_list,
                    start=start,
                    end=end,
                )

            for data in tsd.values():
                if str(data.unit) in BAD_UNITS:
                    data.override_unit("undef")

                filtered = self._filter_timeseries(
                    data,
                    highpass=args.highpass,
                    lowpass=args.lowpass,
                    notch=args.notch,
                )

                if filtered.dtype.kind == "f":  # cast single to double
                    filtered = filtered.astype("float64", order="A", copy=False)

                self.timeseries.append(filtered)

        # report what we have if they asked for it
        self.logger.debug("Channels: %s", self.chan_list)
        self.logger.debug(
            "Start times: %s, duration %s",
            self.start_list,
            self.duration,
        )
        self.logger.debug("Number of time series: %d", len(self.timeseries))

    @staticmethod
    def _filter_timeseries(
        data: TimeSeriesType,
        highpass: QuantityLike | None = None,
        lowpass: QuantityLike | None = None,
        notch: list[QuantityLike] | None = None,
    ) -> TimeSeriesType:
        """Apply highpass, lowpass, and notch filters to some data."""
        # catch nothing to do
        if all(x is None for x in (highpass, lowpass, notch)):
            return data

        # build ZPK
        zpks: list[ZpkType] = []
        if highpass is not None and lowpass is not None:
            zpks.append(filter_design.bandpass(highpass, lowpass, data.sample_rate))
        elif highpass is not None:
            zpks.append(filter_design.highpass(highpass, data.sample_rate))
        elif lowpass is not None:
            zpks.append(filter_design.lowpass(lowpass, data.sample_rate))
        zpks.extend(filter_design.notch(freq, data.sample_rate) for freq in notch or [])
        zpk = filter_design.concatenate_zpks(*zpks)

        # apply forward-backward (zero-phase) filter
        return data.filter(zpk, filtfilt=True)

    # -- plotting --------------------

    def get_xlabel(self) -> str | None:
        """Return default X-axis label for plot."""
        return None

    def get_ylabel(self) -> str | None:
        """Return default Y-axis label for plot."""
        return None

    def get_title(self) -> str:
        """Return default title for plot."""
        highpass = self.args.highpass
        lowpass = self.args.lowpass
        notch = self.args.notch
        filt = ""
        if highpass and lowpass:
            filt += f"band pass ({highpass:.1f}-{lowpass:.1f})"
        elif highpass:
            filt += f"high pass ({highpass:.1f}) "
        elif lowpass:
            filt += f"low pass ({lowpass:.1f}) "
        if notch:
            filt += f", notch ({', '.join(map(str, notch))})"
        return filt

    def get_suptitle(self) -> str:
        """Return default super-title for plot."""
        ts = self.timeseries[0]
        return str(
            ts.name
            or ts.channel
            or "",
        )

    def set_plot_properties(self) -> None:
        """Finalize figure object and show() or save()."""
        self.set_axes_properties()
        self.set_legend()
        self.set_title(self.args.title)
        self.set_suptitle(self.args.suptitle)
        self.set_grid(enable=self.args.nogrid)

    def set_axes_properties(self) -> None:
        """Set properties for each axis (scale, limits, label)."""
        for ax in self.axes:
            self._scale_axes_from_data(ax=ax)
            self._set_xaxis_properties(ax=ax)
            self._set_yaxis_properties(ax=ax)

    @abc.abstractmethod
    def _scale_axes_from_data(self, ax: Axes) -> None:
        """Auto-scale the view based on visible data.

        In this base class, this method does nothing.
        """

    def _set_axis_properties(
        self,
        ax: Axes,
        axis: str,
    ) -> None:
        """Set properties for X/Y axis."""
        axis_name = axis[0]

        def _get(param: str) -> object:
            return getattr(ax, f"get_{axis_name}{param}")()

        def _set(param: str, *args: object, **kwargs) -> None:
            getattr(ax, f"set_{axis_name}{param}")(*args, **kwargs)

        scale = getattr(self.args, f"{axis}scale")
        label = getattr(self.args, f"{axis}label")
        min_ = getattr(self.args, f"{axis}min")
        max_ = getattr(self.args, f"{axis}max")

        # parse limits
        if (
            scale == "auto-gps"
            and min_ is not None
            and max_ is not None
            and max_ < MIN_GPS
        ):
            limits = (min_, min_ + max_)
        else:
            limits = (min_, max_)

        # set limits
        if limits[0] is not None or limits[1] is not None:
            _set("lim", *limits)

        # set scale
        if scale:
            _set("scale", scale)

        # reset scale with epoch if using GPS scale
        if _get("scale") in GPS_SCALES:
            _set("scale", scale, epoch=self.args.epoch)

        # set label
        if label is None:
            label = getattr(self, f"get_{axis_name}label")()
        if label:
            if self.usetex:
                label = label_to_latex(label)
            _set("label", label)

        # log
        dname = axis_name.upper()
        self.logger.debug(
            "%s-axis parameters: scale: %s, limits: %f -> %f",
            dname,
            _get("scale"),
            *cast("tuple[float, float]", _get("lim")),
        )
        self.logger.debug(
            "%s-axis label: %r",
            dname,
            _get("label"),
        )

    def _set_xaxis_properties(self, ax: Axes) -> None:
        """Set properties for X-axis."""
        self._set_axis_properties(ax, "x")

    def _set_yaxis_properties(self, ax: Axes) -> None:
        """Set properties for Y-axis."""
        self._set_axis_properties(ax, "y")

    def set_legend(self, ax: Axes | None = None) -> Legend | None:
        """Create a legend for this product (if applicable)."""
        if ax is None:
            ax = self.ax
        leg = self.ax.legend(prop={"size": 10})
        if leg and self.n_datasets == 1:
            try:
                leg.remove()
            except NotImplementedError:
                leg.set_visible(False)
        return leg

    def set_title(self, title: str, ax: Axes | None = None) -> None:
        """Set the title(s) for these `~matplotlib.axes.Axes`.

        The `Axes.title` actually serves at the sub-title for the plot,
        typically giving processing parameters and information.
        """
        if title is None:
            title_line = self.get_title().rstrip(", ")
        else:
            title_line = title

        if self.usetex:
            title_line = label_to_latex(title_line)
        if title_line:
            (ax or self.ax).set_title(title_line, fontsize=12)
            self.logger.debug("Title is: %s", title_line)

    def set_suptitle(self, suptitle: str | None) -> None:
        """Set the super title for this plot."""
        if not suptitle:
            suptitle = self.get_suptitle()
        if self.usetex:
            suptitle = label_to_latex(suptitle)
        self.plot.suptitle(suptitle, fontsize=18)
        self.logger.debug("Super title is: %s", suptitle)

    def set_grid(
        self,
        ax: Axes | None = None,
        *,
        enable: bool,
    ) -> None:
        """Set the grid parameters for this plot."""
        ax = ax or self.ax

        # Disable
        if not enable:
            ax.grid(visible=False)
            return

        # Major
        ax.grid(
            visible=True,
            which="major",
            color="k",
            linestyle="solid",
        )

        # Minor
        ax.grid(
            visible=True,
            which="minor",
            color="0.06",
            linestyle="dotted",
        )

    def save(self, outfile: str | PathLike | BinaryIO) -> None:
        """Save this product to the target `outfile`."""
        self.plot.savefig(outfile, edgecolor="white", bbox_inches="tight")
        self.logger.info("Wrote %s", outfile)

    def has_more_plots(self) -> bool:
        """Determine whether this product has more plots to be created."""
        return self.plot_num == 0

    @timer
    def _make_plot(self) -> None:
        """Override when product needs multiple saves."""
        self.plot = self.make_plot()

    def _color_by_ifo(self) -> list[str | None]:
        """Return a list of colours to use for datasets.

        In the specific case where each dataset name has a recognised IFO
        prefix, and there is only one of each type, the list will contain
        the relevant colour for that dataset, according to the colour scheme.

        In all other cases a list of `None` is returned.
        """
        # get all name prefices
        names = [
            str(series.name or series.channel).split(":", 1)[0]
            for series in self.timeseries
        ]
        unique = set(names)

        # if not all prefixes are unique, don't do anything
        nnames = len(names)
        if nnames != len(unique):
            return [None] * nnames

        # if not all prefices are in the colour scheme, don't do anything
        if not all(n in GW_OBSERVATORY_COLORS for n in names):
            return [None] * nnames

        # each prefix is in the colour scheme, and is unique, so use the
        # assigned colour
        return [GW_OBSERVATORY_COLORS[n] for n in names]

    # -- the one that does all the work

    def run(self) -> None:
        """Make the plot."""
        self.logger.debug("Arguments:")
        argsd = vars(self.args)
        for key in sorted(argsd):
            self.logger.debug("%15s = %s", key, argsd[key])

        # grab the data
        self.logger.info("---- Loading data ----")
        self.get_data()

        # for each plot
        self.logger.info("---- Generating figures ----")
        show_error = True  # control ours separate from product's
        while self.has_more_plots():
            self._make_plot()
            if self.plot:
                self.set_plot_properties()
                self.add_segs(self.args)
                if self.args.interactive:
                    self.logger.debug(
                        "Interactive manipulation of image should be available.",
                    )
                    self.plot.show()
                else:
                    self.save(self.args.out)
            elif show_error:
                # Some plots reject input data for reasons like all zeroes
                self.logger.warning(
                    "No plot produced because of data validation error.",
                )
                self.got_error = True
                show_error = False
            self.plot_num += 1

        self.logger.info("---- Complete ----")

    def add_segs(self, args: Namespace) -> None:
        """If requested add DQ segments."""
        std_segments = [
            "{ifo}:DMT-GRD_ISC_LOCK_NOMINAL:1",
            "{ifo}:DMT-DC_READOUT_LOCKED:1",
            "{ifo}:DMT-CALIBRATED:1",
            "{ifo}:DMT-ANALYSIS_READY:1",
        ]
        segments = []
        if hasattr(args, "std_seg"):
            if args.std_seg:
                segments = std_segments
            if args.seg:
                for seg in args.seg:
                    # NB: args.seg may be list of lists
                    segments += seg

            chan = args.chan[0][0]
            m = re.match("^([A-Za-z][0-9]):", chan)
            ifo = m.group(1) if m else "?:"

            start: Quantity | None = None
            end = 0
            for ts in self.timeseries:
                if start is not None:
                    start = min(ts.t0, start)
                    end = max(ts.t0 + ts.duration, end)
                else:
                    start = ts.t0
                    end = start + ts.duration

            for segment in segments:
                seg_name = segment.replace("{ifo}", ifo)
                seg_data = DataQualityFlag.query_dqsegdb(
                    seg_name,
                    start,
                    end,
                )

                self.plot.add_segments_bar(seg_data, label=seg_name)


# -- extensions ----------------------

class ImageProduct(CliProduct, metaclass=abc.ABCMeta):
    """Base class for all x/y/color plots."""

    DEFAULT_CMAP: ClassVar[str] = "viridis"
    MAX_DATASETS: ClassVar[int] = 1

    @classmethod
    def init_plot_options(cls, parser: ArgumentParser) -> None:
        """Configure plotting options, including colorbar."""
        super().init_plot_options(parser)
        cls.arg_color_axis(parser)

    @classmethod
    def arg_color_axis(cls, parser: ArgumentParser) -> _ArgumentGroup:
        """Add an `~argparse._ArgumentGroup` for colour-axis options."""
        group = parser.add_argument_group("Colour axis options")
        group.add_argument(
            "--imin",
            type=float,
            help="minimum value for colorbar",
        )
        group.add_argument(
            "--imax",
            type=float,
            help="maximum value for colorbar",
        )
        group.add_argument(
            "--cmap",
            default=cls.DEFAULT_CMAP,
            help=(
                "Colormap. See https://matplotlib.org/examples/color/"
                "colormaps_reference.html for options"
            ),
        )
        group.add_argument(
            "--color-scale",
            choices=("log", "linear"),
            help="scale for colorbar",
        )
        group.add_argument(
            "--norm",
            nargs="?",
            const="median",
            choices=("median", "mean"),
            metavar="NORM",
            help="normalise each pixel against average in that frequency bin",
        )
        group.add_argument(
            "--nocolorbar",
            action="store_true",
            help="hide the colour bar",
        )
        return group

    def get_color_label(self) -> str | None:
        """Return default colorbar label."""
        return None

    def set_axes_properties(self) -> None:
        """Set properties for each axis (scale, limits, label) and create a colorbar."""
        super().set_axes_properties()
        if not self.args.nocolorbar:
            self.set_colorbar()

    def set_colorbar(self) -> None:
        """Create a colorbar for this product."""
        self.ax.colorbar(label=self.get_color_label())

    def set_legend(self, ax: Axes | None = None) -> None:  # noqa: ARG002
        """Create a legend for this plot.

        This method does nothing, since image plots don't have legends.
        """
        return


class FFTMixin:
    """Mixin for `CliProduct` class that will perform FFTs.

    This just adds FFT-based command line options
    """

    DEFAULT_FFTLENGTH: ClassVar[float] = 1.0

    @classmethod
    def init_data_options(cls, parser: ArgumentParser) -> None:
        """Set up data input and signal processing options including FFTs."""
        super().init_data_options(parser)  # type: ignore[misc]
        cls.arg_fft(parser)

    @classmethod
    def arg_fft(cls, parser: ArgumentParser) -> _ArgumentGroup:
        """Add an `~argparse._ArgumentGroup` for FFT options."""
        group = parser.add_argument_group("Fourier transform options")
        group.add_argument(
            "--secpfft",
            type=float,
            default=cls.DEFAULT_FFTLENGTH,
            help="length of FFT in seconds",
        )
        group.add_argument(
            "--overlap",
            type=float,
            help="overlap as fraction of FFT length [0-1)",
        )
        group.add_argument(
            "--window",
            type=str,
            default="hann",
            help="window function to use when overlapping FFTs",
        )
        group.add_argument(
            "--average-method",
            dest="method",
            default=DEFAULT_FFT_METHOD,
            choices=("median", "welch", "bartlett"),
            help="FFT averaging method",
        )
        return group

    @classmethod
    def _arg_faxis(
        cls,
        axis: Literal["x", "y"],
        parser: ArgumentParser,
        **defaults,
    ) -> _ArgumentGroup:
        defaults.setdefault("scale", "log")
        name = f"{axis.title()} axis"
        group = parser.add_argument_group(f"{name} options")

        # label
        group.add_argument(
            f"--{axis}label",
            default=defaults.get("label"),
            dest=f"{axis}label",
            help=f"{name} label",
        )

        # min and max
        for extrema in ("min", "max"):
            meg = group.add_mutually_exclusive_group()
            for ax_ in (axis, "f"):
                meg.add_argument(
                    f"--{ax_}{extrema}",
                    type=float,
                    default=defaults.get(extrema),
                    dest=f"{axis}{extrema}",
                    help=f"{extrema} value for {name}",
                )

        # scale
        scaleg = group.add_mutually_exclusive_group()
        scaleg.add_argument(
            f"--{axis}scale",
            type=str,
            default=defaults.get("scale"),
            dest=f"{axis}scale",
            help=f"scale for {name}",
        )
        for ax_ in (axis, "f"):
            if defaults.get("scale") == "log":
                scaleg.add_argument(
                    f"--nolog{ax_}",
                    action="store_const",
                    dest=f"{axis}scale",
                    const=None,
                    default="log",
                    help=f"use linear {name}",
                )
            else:
                scaleg.add_argument(
                    f"--log{ax_}",
                    action="store_const",
                    dest=f"{axis}scale",
                    const="log",
                    default=None,
                    help="use logarithmic {name}",
                )

        return group

    def _finalize_arguments(self, args: Namespace) -> None:
        if args.overlap is None:
            try:
                args.overlap = recommended_overlap(args.window)
            except ValueError:
                args.overlap = 0.5
        return super()._finalize_arguments(args)  # type: ignore[misc]


class TimeDomainProduct(CliProduct, metaclass=abc.ABCMeta):
    """`CliProduct` with time on the X-axis."""

    @classmethod
    def arg_xaxis(cls, parser: ArgumentParser) -> _ArgumentGroup:
        """Add an `~argparse._ArgumentGroup` for X-axis options.

        This method includes the standard X-axis options, as well as a new
        ``--epoch`` option for the time axis.
        """
        group = super().arg_xaxis(parser)
        group.add_argument(
            "--epoch",
            type=to_gps,
            help=(
                "center X axis on this GPS time, may be absolute date/time or delta"
            ),
        )

        group.add_argument(
            "--std-seg",
            action="store_true",
            help="add DQ segment describing IFO state",
        )
        group.add_argument(
            "--seg",
            type=str,
            nargs="+",
            action="append",
            help="specify one or more DQ segment names",
        )
        return group

    def _finalize_arguments(self, args: Namespace) -> None:
        starts = [float(gps) for gpsl in args.start for gps in gpsl]
        if args.xscale is None:  # set default x-axis scale
            args.xscale = "auto-gps"
        if args.xmin is None:
            args.xmin = min(starts)
        if args.epoch is None:
            args.epoch = args.xmin
        elif args.epoch < MIN_GPS:
            args.epoch += min(starts)

        if args.xmax is None:
            args.xmax = max(starts) + args.duration
        return super()._finalize_arguments(args)

    def get_xlabel(self) -> str:
        """Return default X-axis label for plot."""
        trans = self.ax.xaxis.get_transform()
        if isinstance(trans, GPSTransform):
            epoch = trans.get_epoch()
            unit = trans.get_unit_name()
            utc = re.sub(r"\.0+", "", Time(epoch, format="gps", scale="utc").iso)
            return f"Time ({unit}) from {utc} ({epoch})"
        return ""


class FrequencyDomainProduct(FFTMixin, CliProduct, metaclass=abc.ABCMeta):
    """`CliProduct` with frequency on the X-axis."""

    def get_xlabel(self) -> str:
        """Return default X-axis label for plot."""
        return "Frequency (Hz)"



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""
.. sectionauthor:: Duncan Macleod <duncan.macleod@ligo.org>
.. currentmodule:: gwpy.table

Plotting `EventTable` rate versus time for specific column bins
###############################################################

I would like to study the rate at which event triggers are generated by the
`ExcessPower` gravitational-wave burst detection algorithm, over a small
stretch of data, binned by various thresholds on signal-to-noise ratio (SNR).

The data from which these events were generated contain a simulated
gravitational-wave signal, or hardware injection, used to validate
the performance of the LIGO detectors and downstream data analysis procedures.
"""

# %%
# First, we import the `EventTable` object and read in a set of events from
# a LIGO_LW-format XML file containing a
# :class:`sngl_burst <igwn_ligolw.lsctables.SnglBurstTable>` table
from gwpy.table import EventTable
events = EventTable.read(
    "H1-LDAS_STRAIN-968654552-10.xml.gz",
    tablename="sngl_burst",
    columns=["peak", "snr"],
)

# %%
# .. note::
#
#    Here we manually specify the `columns` to read in order to optimise
#    the `read()` operation to parse only the data we actually need.
#
# Now we can use the :meth:`~EventTable.binned_event_rates` method to
# calculate the event rate in a number of bins of SNR.
rates = events.binned_event_rates(
    1,
    "snr",
    [2, 3, 5, 8],
    operator=">=",
    start=968654552,
    end=968654562,
)

# %%
# .. note::
#
#    The list `[2, 3, 5, 8]` and operator `>=` specifies SNR tresholds of
#    2, 3, 5, and 8.
#
# Finally, we can make a plot:

plot = rates.step()
ax = plot.gca()
ax.set_xlim(968654552, 968654562)
ax.set_ylabel("Event rate [Hz]")
ax.set_title("LIGO Hanford Observatory event rate for HW100916")
ax.legend(title="SNR $>=$")
plot.show()



# Copyright (c) 2013-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.signal.spectral._ui`."""

import numpy
import pytest
from astropy import units
from scipy import signal

from ...testing.utils import (
    assert_array_equal,
    assert_quantity_sub_equal,
)
from ...timeseries import TimeSeries
from ..spectral import _ui as fft_ui


@pytest.mark.parametrize(("secs", "rate", "result"), [
    # all floats
    (4, 256, 1024),
    # duration as a quantity
    (1 * units.minute, 16, 960),
    # duration and rate as quantities
    (4 * units.second, 16.384 * units.kiloHertz, 65536),
])
def test_seconds_to_samples(secs, rate, result):
    """Test :func:`gwpy.signal.spectral.ui.seconds_to_samples`."""
    assert fft_ui.seconds_to_samples(secs, rate) == result


def test_normalize_fft_params():
    """Test :func:`gwpy.signal.spectral.ui.normalize_fft_params`."""
    assert fft_ui.normalize_fft_params(
        TimeSeries(numpy.zeros(1024), sample_rate=256),
    ) == {
        "nfft": 1024,
        "noverlap": 0,
    }


def test_normalize_fft_params_window_str():
    ftp = fft_ui.normalize_fft_params(
        TimeSeries(numpy.zeros(1024), sample_rate=256),
        {"window": "hann"},
    )
    win = signal.get_window("hann", 1024)
    assert ftp.pop("nfft") == 1024
    assert ftp.pop("noverlap") == 512
    assert_array_equal(ftp.pop("window"), win)
    assert not ftp


def test_normalize_fft_params_window_array():
    win = signal.get_window(("kaiser", 14), 1024)
    ftp = fft_ui.normalize_fft_params(
        TimeSeries(numpy.zeros(1024), sample_rate=256),
        {"window": win},
    )
    assert_array_equal(ftp.pop("window"), win)


@pytest.mark.requires("lal")
@pytest.mark.parametrize("win", [
    "hann",
    signal.get_window(("kaiser", 14), 1024),
])
def test_normalize_fft_params_window_lal(win):
    import lal

    from gwpy.signal.spectral._lal import welch

    ftp = fft_ui.normalize_fft_params(
        TimeSeries(numpy.zeros(1024, dtype="float32"), sample_rate=256),
        kwargs={"window": win},
        func=welch,
    )
    assert isinstance(ftp.pop("window"), lal.REAL4Window)


def test_chunk_timeseries():
    """Test :func:`gwpy.signal.spectral.ui._chunk_timeseries`."""
    a = TimeSeries(numpy.arange(400))
    chunks = list(fft_ui._chunk_timeseries(a, 100, 50))
    for i, (idxa, idxb) in enumerate([
        (None, 150),
        (75, 225),
        (175, 325),
        (250, 400),
    ]):
        assert_quantity_sub_equal(chunks[i], a[idxa:idxb])



# Copyright (c) 2019-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Extensions of `~matplotlib.legend` for gwpy."""

from __future__ import annotations

from typing import TYPE_CHECKING

from matplotlib import legend_handler

if TYPE_CHECKING:
    from collections.abc import Sequence

    from matplotlib.artist import Artist
    from matplotlib.legend import Legend
    from matplotlib.transforms import Transform

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


class HandlerLine2D(legend_handler.HandlerLine2D):
    """Custom Line2D legend handler that draws lines with custom linewidth.

    Parameters
    ----------
    linewidth : `float`, optional
        the linewidth to use when drawing lines on the legend

    kwargs
        All keyword arguments are passed to
        :class`matplotlib.legend_handler.HandlerLine2D`.

    See Also
    --------
    matplotlib.legend_handler.HanderLine2D
        For all information.
    """

    def __init__(self, linewidth: float = 6., **kwargs) -> None:
        """Initialise a new `HandlerLine2D`."""
        super().__init__(**kwargs)
        self._linewidth = linewidth

    def create_artists(
        self,
        legend: Legend,
        orig_handle: Artist,
        xdescent: float,
        ydescent: float,
        width: float,
        height: float,
        fontsize: float,
        trans: Transform,
    ) -> Sequence[Artist]:
        """Create artists for this legend handler."""
        artists = super().create_artists(
            legend,
            orig_handle,
            xdescent,
            ydescent,
            width,
            height,
            fontsize,
            trans,
        )
        artists[0].set_linewidth(self._linewidth)
        return artists



# Copyright (c) 2018-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.cli.timeseries`."""

from __future__ import annotations

from ... import cli
from .base import (
    _TestTimeDomainProduct,
    update_namespace,
)


class TestTimeSeriesProduct(_TestTimeDomainProduct[cli.TimeSeriesProduct]):
    """Tests for `gwpy.cli.TimeSeriesProduct`."""

    TEST_CLASS = cli.TimeSeriesProduct
    ACTION = "timeseries"

    def test_get_title(self, prod: cli.TimeSeriesProduct):
        """Test `TimeSeriesProduct.get_title`."""
        update_namespace(prod.args, highpass=10, lowpass=100)
        t = f"Fs: (), duration: {prod.args.duration}, band pass (10.0-100.0)"
        assert prod.get_title() == t

    def test_get_suptitle(self, prod: cli.TimeSeriesProduct):
        """Test `TimeSeriesProduct.get_suptitle`."""
        assert prod.get_suptitle() == f"Time series: {prod.chan_list[0]}"



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.cli.coherencegram`."""

from ... import cli
from .test_coherence import TestCoherenceProduct as _TestCoherenceProduct
from .test_spectrogram import TestSpectrogramProduct as _TestSpectrogramProduct


class TestCoherencegramProduct(_TestSpectrogramProduct):
    """Tests for `CoherencegramProduct`."""

    TEST_CLASS = cli.CoherencegramProduct
    ACTION = "coherencegram"
    TEST_ARGS = _TestCoherenceProduct.TEST_ARGS

    def test_finalize_arguments(self, prod):
        """Test that `CoherencegramProduct.finalize_arguments` sets defaults."""
        assert prod.args.cmap == "plasma"
        assert prod.args.color_scale == "linear"
        assert prod.args.imin == 0.
        assert prod.args.imax == 1.

    def test_get_suptitle(self, prod):
        """Test `CoherencegramProduct.get_suptitle()`."""
        assert prod.get_suptitle() == (
            "Coherence spectrogram: "
            f"{prod.chan_list[0]} vs {prod.chan_list[1]}"
        )



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""GPS axis locators, formatters, and scales."""

from __future__ import annotations

import contextlib
from decimal import Decimal
from numbers import Number
from typing import (
    TYPE_CHECKING,
    cast,
)

import numpy
from astropy import units
from matplotlib import ticker
from matplotlib.scale import (
    LinearScale,
    _get_scale_docs as get_scale_docs,
    get_scale_names,
    register_scale,
)
from matplotlib.transforms import Transform

try:
    from matplotlib import _docstring
except ImportError:  # maybe matplotlib >= 3.9?
    _docstring = None  # type: ignore[assignment]

from ..time import (
    from_gps,
    to_gps,
)

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import (
        Any,
        Literal,
    )

    from astropy.units import UnitBase
    from matplotlib.axis import Axis
    from numpy.typing import ArrayLike

    from ..time import SupportsToGps
    from ..typing import UnitLike

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

#: Maximum number of week ticks to display
WEEK_SCALE_MAJOR_TICKS = 6

#: Supported time scales
TIME_UNITS = (
    units.nanosecond,
    units.microsecond,
    units.millisecond,
    units.second,
    units.minute,
    units.hour,
    units.day,
    units.week,
    units.year,
    units.kiloyear,
    units.megayear,
    units.gigayear,
)

GPS_SCALES = {}


def _truncate(f: float, n: int) -> str:
    """Truncates/pads a float `f` to `n` decimal places without rounding.

    From https://stackoverflow.com/a/783927/1307974 (CC-BY-SA)
    """
    s = str(f)
    if "e" in s or "E" in s:
        return f"{f:.{n}f}"
    i, p, d = s.partition(".")
    return ".".join([i, (d+"0"*n)[:n]])


# -- base mixin for all GPS manipulations

class GPSMixin:
    """Mixin adding GPS-related attributes to any class."""

    def __init__(
        self,
        *args: Any,  # noqa: ANN401
        unit: UnitBase | None = None,
        epoch: Number | Decimal | SupportsToGps | None = None,
        **kwargs,
    ) -> None:
        """Initialise a new GPS-scaled object."""
        self.set_unit(unit)
        self.set_epoch(epoch)

        # call super for __init__ if this is part of a larger MRO
        with contextlib.suppress(TypeError):
            super().__init__(*args, **kwargs)

    def get_epoch(self) -> float | None:
        """Return the GPS epoch."""
        return self._epoch

    def set_epoch(self, epoch: Number | Decimal | SupportsToGps | None) -> None:
        """Set the GPS epoch."""
        if epoch is None:
            self._epoch = None
        else:
            self._epoch = float(to_gps(epoch))

    epoch = property(
        fget=get_epoch,
        fset=set_epoch,
        doc=get_epoch.__doc__,
    )

    def get_unit(self) -> UnitBase | None:
        """GPS step scale."""
        return self._unit

    def set_unit(self, unit: UnitLike | Number | None) -> None:
        """Set the GPS step scale."""
        # accept all core time units
        if (
            unit is None
            or (isinstance(unit, units.NamedUnit) and unit.physical_type == "time")
        ):
            self._unit = unit
            return

        # convert float to custom unit in seconds
        if isinstance(unit, Number):
            unit = units.Unit(unit * units.second)

        # otherwise, should be able to convert to a time unit
        try:
            unit = units.Unit(unit)
        except ValueError:
            # catch annoying plurals
            unit = units.Unit(str(unit).rstrip("s"))
        unit = cast("UnitBase", unit)

        # decompose and check that it's actually a time unit
        dec = unit.decompose()
        if dec.bases != [units.second]:
            msg = f"cannot set GPS unit to '{unit}'"
            raise ValueError(msg)

        # check equivalent units
        for other in TIME_UNITS:
            if other.decompose().scale == dec.scale:
                self._unit = other
                return

        msg = f"unrecognised unit '{unit}'"
        raise ValueError(msg)

    unit = property(
        fget=get_unit,
        fset=set_unit,
        doc=get_unit.__doc__,
    )

    def get_unit_name(self) -> str | None:
        """Return the name of the unit for this GPS scale.

        Note that this returns a simply-pluralised version of the name.
        """
        if not self.unit:
            return None
        try:
            name = self.unit.long_names[0]
        except IndexError:
            name = self.unit.name
        if len(name) > 1:
            return name + "s"  # pluralise for humans
        return name

    def get_scale(self) -> float:
        """Return the scale (in seconds) of the current GPS unit."""
        if self.unit is None:
            return 1
        return self.unit.decompose().scale

    scale = property(fget=get_scale, doc=get_scale.__doc__)


# -- GPS transforms ----------------

class _GPSTransformBase(GPSMixin, Transform):
    """Transform GPS time into N * scale from epoch.

    This class uses the `decimal.Decimal` object to protect against precision
    errors when converting to and from GPS times that may have 19 significant
    digits, which is more than `float` can handle.

    There is some logic to _only_ use the slow decimal transforms when
    absolutely necessary, normally when transforming tick positions.
    """

    input_dims = 1
    output_dims = 1
    is_separable = True
    is_affine = True
    has_inverse = True

    def transform(self, values: ArrayLike) -> numpy.ndarray:
        """Transform an array of GPS times."""
        # format ticks using decimal for precision display
        if isinstance(values, Number | Decimal):
            return self._transform_decimal(values, self.epoch or 0, self.scale)
        return super().transform(values)

    def transform_non_affine(self, values: ArrayLike) -> numpy.ndarray:
        """Transform an array of GPS times.

        This method is designed to filter out transformations that will
        generate text elements that require exact precision, and use
        `Decimal` objects to do the transformation, and simple `float`
        otherwise.
        """
        scale = self.scale or 1
        epoch = self.epoch or 0

        values = numpy.asarray(values)

        # handle simple or data transformations with floats
        if self._parents or (  # part of composite transform (from draw())
            # no large additions
            epoch == 0
            # no multiplications
            and scale == 1
        ):
            return self._transform(values, float(epoch), float(scale))

        # otherwise do things carefully (and slowly) with Decimals
        # -- ideally this only gets called for transforming tick positions
        flat = values.flatten()

        def _trans(x: float) -> float:
            return self._transform_decimal(x, epoch, scale)

        return numpy.fromiter(
            map(_trans, flat),
            dtype=float,
            count=flat.size,
        ).reshape(values.shape)

    @staticmethod
    def _transform(
        value: numpy.ndarray,
        epoch: float,
        scale: float,
    ) -> numpy.ndarray:
        """Transform the GPS ``value`` into a scaled time relative to an epoch."""
        # convert GPS into scaled time from epoch
        return (value - epoch) / scale

    @classmethod
    def _transform_decimal(
        cls,
        value: float,
        epoch: float,
        scale: float,
    ) -> float:
        """Transform to/from GPS using `decimal.Decimal` for precision."""
        vdec = Decimal(_truncate(value, 12))
        edec = Decimal(_truncate(epoch, 12))
        sdec = Decimal(_truncate(scale, 12))
        return type(value)(cls._transform(vdec, edec, sdec))  # type: ignore[arg-type]


class GPSTransform(_GPSTransformBase):
    """Transform GPS into time (scaled units) from epoch."""

    def inverted(self) -> InvertedGPSTransform:
        """Return the inverse of this `GPSTransform`."""
        return InvertedGPSTransform(unit=self.unit, epoch=self.epoch)


class InvertedGPSTransform(_GPSTransformBase):
    """Transform time (scaled units) from epoch into GPS time."""

    @staticmethod
    def _transform(
        value: numpy.ndarray,
        epoch: float,
        scale: float,
    ) -> numpy.ndarray:
        """Transform the scaled time back into GPS."""
        return value * scale + epoch

    def inverted(self) -> GPSTransform:
        """Return the inverse of this `InvertedGPSTransform`."""
        return GPSTransform(unit=self.unit, epoch=self.epoch)


# -- locators and formatters ---------

class GPSAutoLocator(ticker.MaxNLocator):
    """Find the best position for ticks on a given axis from the data.

    This auto-locator gives a simple extension to the matplotlib
    `~matplotlib.ticker.AutoLocator` allowing for variations in scale
    and zero-time epoch.
    """

    def __init__(
        self,
        nbins: int | Literal["auto"] | None = 12,
        **kwargs,
    ) -> None:
        """Initialise a new `GPSAutoLocator`.

        Each of the `epoch` and `scale` keyword arguments should match those
        passed to the `GPSFormatter`
        """
        super().__init__(
            nbins=nbins,
            **kwargs,
        )

    def tick_values(self, vmin: float, vmax: float) -> Sequence[float]:
        """Generate the list of tick values for the given interval."""
        self.axis: Axis
        transform = self.axis.get_transform()
        unit = transform.get_unit()
        steps = self._steps

        vmin, vmax = transform.transform((vmin, vmax))

        # if less than 6 weeks, major tick every week
        if (
            steps is None
            and unit == units.week
            and vmax - vmin <= WEEK_SCALE_MAJOR_TICKS
        ):
            self.set_params(steps=[1, 10])
        else:
            self.set_params(steps=None)

        try:
            ticks = super().tick_values(vmin, vmax)
        finally:
            self._steps = steps
        return transform.inverted().transform(ticks)


class GPSAutoMinorLocator(ticker.AutoMinorLocator):
    """Find the best position for minor ticks on a given GPS-scaled axis."""

    def __call__(self) -> Sequence[float]:
        """Return the locations of the ticks."""
        self.axis: Axis
        majorlocs = self.axis.get_majorticklocs()
        trans = self.axis.get_transform()
        try:
            majorstep = majorlocs[1] - majorlocs[0]
        except IndexError:
            # Need at least two major ticks to find minor tick locations
            # TODO (@duncanmmacleod): Figure out a way to still be able to
            # display minor ticks without two major ticks visible.
            # For now, just display no ticks at all.
            majorstep = 0

        if self.ndivs is None:
            if majorstep == 0:
                # TODO (@duncanmmacleod): Need a better way to figure out ndivs
                ndivs = 1
            else:
                scale_ = trans.get_scale()
                gpsstep = majorstep / scale_
                x = round(10 ** (numpy.log10(gpsstep) % 1))
                if trans.unit == units.week and gpsstep == 1:
                    ndivs = 7
                elif trans.unit == units.year and gpsstep <= 1:
                    ndivs = 6
                elif trans.unit != units.day and x in [1, 5, 10]:
                    ndivs = 5
                else:
                    ndivs = 4
        else:
            ndivs = self.ndivs

        minorstep = majorstep / ndivs

        vmin, vmax = self.axis.get_view_interval()
        if vmin > vmax:
            vmin, vmax = vmax, vmin

        if numpy.size(majorlocs):
            epoch = majorlocs[0]
            tmin = numpy.floor((vmin - epoch) / minorstep) * minorstep
            tmax = numpy.ceil((vmax - epoch) / minorstep) * minorstep
            locs = numpy.arange(tmin, tmax, minorstep) + epoch
            cond = numpy.abs((locs - epoch) % majorstep) > minorstep / 10.0
            locs = locs.compress(cond)
        else:
            locs = []

        return self.raise_if_exceeds(numpy.array(locs))


class GPSFormatter(ticker.Formatter):
    """Format ticks on a `GPSScale` axis."""

    def __call__(
        self,
        t: float,
        pos: int | None = None,  # noqa: ARG002
    ) -> str:
        """Format a tick on this scale."""
        trans = self.axis.get_transform()
        flt = float(trans.transform(float(t)))
        if flt.is_integer():
            return str(int(flt))
        return str(flt)


# -- scales --------------------------

class GPSScale(GPSMixin, LinearScale):
    """A GPS scale, displaying time (scaled units) from an epoch.

    Parameters
    ----------
    axis : `matplotlib.axis.Axis`.
        The axis to scale.

    unit : `astropy.units.Unit`, optional
        The unit to use for ticks on the axis.

    epoch : `float`, `gwpy.time.LIGOTimeGPS`, optional
        The GPS epoch (origin) for axis ticks.
    """

    name = "auto-gps"
    Transform = GPSTransform
    InvertedTransform = InvertedGPSTransform

    def __init__(
        self,
        axis: Axis,
        unit: UnitBase | None = None,
        epoch: Number | Decimal | SupportsToGps | None = None,
    ) -> None:
        """Initialise this `GPSScale`."""
        super().__init__(unit=unit, epoch=epoch)
        self.axis = axis
        # set tight scaling on parent axes
        getattr(axis.axes, f"set_{axis.axis_name}margin")(0)

    def set_default_locators_and_formatters(self, axis: Axis) -> None:
        """Set the defualt locators and formatters for ``axis``."""
        axis.set_major_locator(GPSAutoLocator())
        axis.set_major_formatter(GPSFormatter())
        axis.set_minor_locator(GPSAutoMinorLocator())
        axis.set_minor_formatter(ticker.NullFormatter())

    @staticmethod
    def _lim(axis: Axis) -> tuple[float, float]:
        """Find the current view limits of this ``axis``."""
        # if autoscaling and datalim is set, use it
        dlim = axis.get_data_interval()
        if (
            getattr(axis.axes, f"get_autoscale{axis.axis_name}_on")()
            and not numpy.isinf(dlim).any()
        ):
            return dlim

        # otherwise use the view lim
        return axis.get_view_interval()

    def _auto_epoch(self, axis: Axis) -> int:
        """Find the best GPS epoch (origin) for this ``axis``."""
        # use the lower data/view limit as the epoch
        epoch = round(self._lim(axis)[0])

        # round epoch in successive units for large scales
        unit = self.get_unit()
        date = from_gps(epoch)
        fields = ("second", "minute", "hour", "day")
        for i, u in enumerate(fields[1:]):
            if unit < units.Unit(u):
                break
            if u in ("day",):
                date = date.replace(**{fields[i]: 1})
            else:
                date = date.replace(**{fields[i]: 0})
        return int(to_gps(date))

    def _auto_unit(self, axis: Axis) -> UnitBase:
        """Find the best scaled unit for this ``axis``."""
        # get width of axis
        vmin, vmax = self._lim(axis)
        duration = vmax - vmin

        # find time unit that fits the duration well;
        # the magic scaling of 4 or 0.01 is entirely arbitrary,
        # but in practice results in figures that scale nicely
        for scale in TIME_UNITS[::-1]:
            base = scale.decompose().scale
            # for large durations, prefer smaller units
            if scale > units.second:
                base *= 4
            # for smaller durations, prefer larger units
            else:
                base *= 0.01
            if duration >= base:
                return scale

        # if nothing else worked, just use seconds
        return units.second

    def get_transform(self) -> GPSTransform:
        """Return the `GPSTransform` associated with this scale."""
        # get current settings
        epoch = self.get_epoch()
        unit = self.get_unit()

        # dynamically set epoch and/or unit if None
        if unit is None:
            self.set_unit(self._auto_unit(self.axis))
        if epoch is None:
            self.set_epoch(self._auto_epoch(self.axis))

        # build transform on-the-fly
        try:
            return self.Transform(
                unit=self.get_unit(),
                epoch=self.get_epoch(),
            )
        finally:  # reset to current settings
            self.set_epoch(epoch)
            self.set_unit(unit)


# -- registrations -------------------

def register_gps_scale(scale_class: type[GPSScale]) -> None:
    """Register a new GPS scale.

    ``scale_class`` must be a subclass of `GPSScale`.
    """
    register_scale(scale_class)
    GPS_SCALES[scale_class.name] = scale_class


def _gps_scale_factory(unit: UnitBase) -> type[GPSScale]:
    """Construct a GPSScale for this unit."""

    class FixedGPSScale(GPSScale):
        """`GPSScale` for a specific GPS time unit."""

        name = (unit.long_names or unit.names)[0] + "s"

        def __init__(
            self,
            axis: Axis,
            epoch: Number | Decimal | SupportsToGps | None = None,
        ) -> None:
            super().__init__(axis, epoch=epoch, unit=unit)

    return FixedGPSScale


register_gps_scale(GPSScale)  # auto-gps

for _unit in TIME_UNITS:
    # don't go past 'year' for GPSScale
    if _unit is units.kiloyear:
        break
    register_gps_scale(_gps_scale_factory(_unit))

# update the docstring for matplotlib scale methods
with contextlib.suppress(AttributeError):
    _docstring.interpd.params.update(
        scale=" | ".join([repr(x) for x in get_scale_names()]),
        scale_docs=get_scale_docs().rstrip(),
    )



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.io.gwf`."""

import pytest

from ...testing.utils import TEST_GWF_FILE
from .. import gwf as io_gwf

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def test_open_gwf_r():
    """Test `open_gwf()` for reading."""
    mod_framecpp = pytest.importorskip("LDAStools.frameCPP")
    open_gwf = io_gwf.get_backend_function("open_gwf", backend="frameCPP")
    assert isinstance(open_gwf(TEST_GWF_FILE), mod_framecpp.IFrameFStream)


def test_open_gwf_w(tmp_path):
    """Test `open_gwf()` for writing."""
    mod_framecpp = pytest.importorskip("LDAStools.frameCPP")
    open_gwf = io_gwf.get_backend_function("open_gwf", backend="frameCPP")
    tmp = tmp_path / "test.gwf"
    assert isinstance(open_gwf(tmp, mode="w"), mod_framecpp.OFrameFStream)


def test_open_gwf_w_file_url(tmp_path):
    """Test `open_gwf()` for writing with a file:// URL."""
    mod_framecpp = pytest.importorskip("LDAStools.frameCPP")
    open_gwf = io_gwf.get_backend_function("open_gwf", backend="frameCPP")
    # check that we can use a file:// URL as well
    tmp = tmp_path / "test.gwf"
    assert isinstance(
        open_gwf(tmp.as_uri(), mode="w"),
        mod_framecpp.OFrameFStream,
    )


@pytest.mark.requires("LDAStools.frameCPP")
def test_open_gwf_a_error():
    """Test that `open_gwf()` raises an error for append mode."""
    open_gwf = io_gwf.get_backend_function("open_gwf", backend="frameCPP")
    with pytest.raises(
        ValueError,
        match="mode must be either 'r' or 'w'",
    ):
        open_gwf("test", mode="a")


@pytest.mark.requires("LDAStools.frameCPP")
def test_create_frvect(noisy_sinusoid):
    """Test `create_frvect()`."""
    create_frvect = io_gwf.get_backend_function(
        "create_frvect",
        backend="frameCPP",
    )
    vect = create_frvect(noisy_sinusoid)
    assert vect.nData == noisy_sinusoid.size
    assert vect.nBytes == noisy_sinusoid.nbytes
    assert vect.name == noisy_sinusoid.name
    assert vect.unitY == noisy_sinusoid.unit
    xdim = vect.GetDim(0)
    assert xdim.unitX == noisy_sinusoid.xunit
    assert xdim.dx == noisy_sinusoid.dx.value
    assert xdim.startX == noisy_sinusoid.x0.value
